(function (global, factory) {
	typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('react')) :
	typeof define === 'function' && define.amd ? define(['exports', 'react'], factory) :
	(factory((global.ReactValidationLayer = global.ReactValidationLayer || {}),global.React));
}(this, (function (exports,React) { 'use strict';

var React__default = 'default' in React ? React['default'] : React;

var Constant = Object.freeze({
  DEFAULT_LAYER_ID: 'form',
  DEFAULT_DEBOUNCE_INTERVAL: 700,
  FIELD_ID_DOM_DATA_ATTRIBUTE: 'data-fieldid',
  FIELD_ID_DELIMITER: '.',
  FIELD_ID_COLLECTION_INDEX_DELIMITER: '/',
  FIELD_STATE_ID_DELIMITER: '---',
  FIELD_DOM_ID_DELIMITER: '___',
  FIELD_PROPS_STATE_ID_PREFIX: 'fieldPropsState',
  FIELD_VALIDATION_STATE_ID_PREFIX: 'fieldValidationState',
  COLLECTION_DATA_TYPE: 'collection'
});

var DefaultStatus = Object.freeze({
  SUCCESS: 'success',
  FAILURE: 'failure'
});

var Strategy = Object.freeze({
  DEFAULT: 'onFirstSuccessOrFirstBlur',

  ON_MOUNT: 'onMount',
  ON_FIRST_CHANGE: 'onFirstChange',
  ON_FIRST_BLUR: 'onFirstBlur',
  ON_FIRST_SUCCESS: 'onFirstSuccess',
  ON_FIRST_SUCCESS_OR_FIRST_BLUR: 'onFirstSuccessOrFirstBlur',
  ON_FIRST_SUBMIT: 'onFirstSubmit'
});

var AsyncStrategy = Object.freeze({
  DEFAULT: 'onChange',

  ON_BLUR: 'onBlur',
  ON_CHANGE: 'onChange'
});

var DebounceStatus = Object.freeze({
  INITIALIZED: 'initialized',
  DEBOUNCED: 'debounced'
});

/**
 * @desc Builds multiline error message with debug info.
 *
 */


function buildErrorMessage(_ref) {
  var layerId = _ref.layerId,
      fieldId = _ref.fieldId,
      message = _ref.message;

  var normalizedMessage = Array.isArray(message) ? message : [message];
  var isDefaultLayerId = layerId === Constant.DEFAULT_LAYER_ID;
  var layerIdOutput = layerId && !isDefaultLayerId ? '    Layer: ' + layerId : null;
  var fieldIdOutput = fieldId ? '    Field: ' + fieldId : null;
  var debugInfoTitle = layerIdOutput || fieldIdOutput ? '--- Debug info:' : null;

  var errorMessage = normalizedMessage.concat(debugInfoTitle, layerIdOutput, fieldIdOutput).filter(function (v) {
    return v;
  }).join('\n');

  return '[react-validation-layer]: ' + errorMessage;
}

var commonjsGlobal = typeof window !== 'undefined' ? window : typeof global !== 'undefined' ? global : typeof self !== 'undefined' ? self : {};





function createCommonjsModule(fn, module) {
	return module = { exports: {} }, fn(module, module.exports), module.exports;
}

/**
 * lodash (Custom Build) <https://lodash.com/>
 * Build: `lodash modularize exports="npm" -o ./`
 * Copyright jQuery Foundation and other contributors <https://jquery.org/>
 * Released under MIT license <https://lodash.com/license>
 * Based on Underscore.js 1.8.3 <http://underscorejs.org/LICENSE>
 * Copyright Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
 */

/** Used as the `TypeError` message for "Functions" methods. */
var FUNC_ERROR_TEXT = 'Expected a function';

/** Used to stand-in for `undefined` hash values. */
var HASH_UNDEFINED = '__lodash_hash_undefined__';

/** Used as references for various `Number` constants. */
var INFINITY = 1 / 0;
var MAX_SAFE_INTEGER = 9007199254740991;

/** `Object#toString` result references. */
var funcTag = '[object Function]';
var genTag = '[object GeneratorFunction]';
var symbolTag = '[object Symbol]';

/** Used to match property names within property paths. */
var reIsDeepProp = /\.|\[(?:[^[\]]*|(["'])(?:(?!\1)[^\\]|\\.)*?\1)\]/;
var reIsPlainProp = /^\w*$/;
var reLeadingDot = /^\./;
var rePropName = /[^.[\]]+|\[(?:(-?\d+(?:\.\d+)?)|(["'])((?:(?!\2)[^\\]|\\.)*?)\2)\]|(?=(?:\.|\[\])(?:\.|\[\]|$))/g;

/**
 * Used to match `RegExp`
 * [syntax characters](http://ecma-international.org/ecma-262/7.0/#sec-patterns).
 */
var reRegExpChar = /[\\^$.*+?()[\]{}|]/g;

/** Used to match backslashes in property paths. */
var reEscapeChar = /\\(\\)?/g;

/** Used to detect host constructors (Safari). */
var reIsHostCtor = /^\[object .+?Constructor\]$/;

/** Used to detect unsigned integer values. */
var reIsUint = /^(?:0|[1-9]\d*)$/;

/** Detect free variable `global` from Node.js. */
var freeGlobal = typeof commonjsGlobal == 'object' && commonjsGlobal && commonjsGlobal.Object === Object && commonjsGlobal;

/** Detect free variable `self`. */
var freeSelf = typeof self == 'object' && self && self.Object === Object && self;

/** Used as a reference to the global object. */
var root = freeGlobal || freeSelf || Function('return this')();

/**
 * Gets the value at `key` of `object`.
 *
 * @private
 * @param {Object} [object] The object to query.
 * @param {string} key The key of the property to get.
 * @returns {*} Returns the property value.
 */
function getValue(object, key) {
  return object == null ? undefined : object[key];
}

/**
 * Checks if `value` is a host object in IE < 9.
 *
 * @private
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is a host object, else `false`.
 */
function isHostObject(value) {
  // Many host objects are `Object` objects that can coerce to strings
  // despite having improperly defined `toString` methods.
  var result = false;
  if (value != null && typeof value.toString != 'function') {
    try {
      result = !!(value + '');
    } catch (e) {}
  }
  return result;
}

/** Used for built-in method references. */
var arrayProto = Array.prototype;
var funcProto = Function.prototype;
var objectProto = Object.prototype;

/** Used to detect overreaching core-js shims. */
var coreJsData = root['__core-js_shared__'];

/** Used to detect methods masquerading as native. */
var maskSrcKey = (function() {
  var uid = /[^.]+$/.exec(coreJsData && coreJsData.keys && coreJsData.keys.IE_PROTO || '');
  return uid ? ('Symbol(src)_1.' + uid) : '';
}());

/** Used to resolve the decompiled source of functions. */
var funcToString = funcProto.toString;

/** Used to check objects for own properties. */
var hasOwnProperty = objectProto.hasOwnProperty;

/**
 * Used to resolve the
 * [`toStringTag`](http://ecma-international.org/ecma-262/7.0/#sec-object.prototype.tostring)
 * of values.
 */
var objectToString = objectProto.toString;

/** Used to detect if a method is native. */
var reIsNative = RegExp('^' +
  funcToString.call(hasOwnProperty).replace(reRegExpChar, '\\$&')
  .replace(/hasOwnProperty|(function).*?(?=\\\()| for .+?(?=\\\])/g, '$1.*?') + '$'
);

/** Built-in value references. */
var Symbol$1 = root.Symbol;
var splice = arrayProto.splice;

/* Built-in method references that are verified to be native. */
var Map = getNative(root, 'Map');
var nativeCreate = getNative(Object, 'create');

/** Used to convert symbols to primitives and strings. */
var symbolProto = Symbol$1 ? Symbol$1.prototype : undefined;
var symbolToString = symbolProto ? symbolProto.toString : undefined;

/**
 * Creates a hash object.
 *
 * @private
 * @constructor
 * @param {Array} [entries] The key-value pairs to cache.
 */
function Hash(entries) {
  var index = -1,
      length = entries ? entries.length : 0;

  this.clear();
  while (++index < length) {
    var entry = entries[index];
    this.set(entry[0], entry[1]);
  }
}

/**
 * Removes all key-value entries from the hash.
 *
 * @private
 * @name clear
 * @memberOf Hash
 */
function hashClear() {
  this.__data__ = nativeCreate ? nativeCreate(null) : {};
}

/**
 * Removes `key` and its value from the hash.
 *
 * @private
 * @name delete
 * @memberOf Hash
 * @param {Object} hash The hash to modify.
 * @param {string} key The key of the value to remove.
 * @returns {boolean} Returns `true` if the entry was removed, else `false`.
 */
function hashDelete(key) {
  return this.has(key) && delete this.__data__[key];
}

/**
 * Gets the hash value for `key`.
 *
 * @private
 * @name get
 * @memberOf Hash
 * @param {string} key The key of the value to get.
 * @returns {*} Returns the entry value.
 */
function hashGet(key) {
  var data = this.__data__;
  if (nativeCreate) {
    var result = data[key];
    return result === HASH_UNDEFINED ? undefined : result;
  }
  return hasOwnProperty.call(data, key) ? data[key] : undefined;
}

/**
 * Checks if a hash value for `key` exists.
 *
 * @private
 * @name has
 * @memberOf Hash
 * @param {string} key The key of the entry to check.
 * @returns {boolean} Returns `true` if an entry for `key` exists, else `false`.
 */
function hashHas(key) {
  var data = this.__data__;
  return nativeCreate ? data[key] !== undefined : hasOwnProperty.call(data, key);
}

/**
 * Sets the hash `key` to `value`.
 *
 * @private
 * @name set
 * @memberOf Hash
 * @param {string} key The key of the value to set.
 * @param {*} value The value to set.
 * @returns {Object} Returns the hash instance.
 */
function hashSet(key, value) {
  var data = this.__data__;
  data[key] = (nativeCreate && value === undefined) ? HASH_UNDEFINED : value;
  return this;
}

// Add methods to `Hash`.
Hash.prototype.clear = hashClear;
Hash.prototype['delete'] = hashDelete;
Hash.prototype.get = hashGet;
Hash.prototype.has = hashHas;
Hash.prototype.set = hashSet;

/**
 * Creates an list cache object.
 *
 * @private
 * @constructor
 * @param {Array} [entries] The key-value pairs to cache.
 */
function ListCache(entries) {
  var index = -1,
      length = entries ? entries.length : 0;

  this.clear();
  while (++index < length) {
    var entry = entries[index];
    this.set(entry[0], entry[1]);
  }
}

/**
 * Removes all key-value entries from the list cache.
 *
 * @private
 * @name clear
 * @memberOf ListCache
 */
function listCacheClear() {
  this.__data__ = [];
}

/**
 * Removes `key` and its value from the list cache.
 *
 * @private
 * @name delete
 * @memberOf ListCache
 * @param {string} key The key of the value to remove.
 * @returns {boolean} Returns `true` if the entry was removed, else `false`.
 */
function listCacheDelete(key) {
  var data = this.__data__,
      index = assocIndexOf(data, key);

  if (index < 0) {
    return false;
  }
  var lastIndex = data.length - 1;
  if (index == lastIndex) {
    data.pop();
  } else {
    splice.call(data, index, 1);
  }
  return true;
}

/**
 * Gets the list cache value for `key`.
 *
 * @private
 * @name get
 * @memberOf ListCache
 * @param {string} key The key of the value to get.
 * @returns {*} Returns the entry value.
 */
function listCacheGet(key) {
  var data = this.__data__,
      index = assocIndexOf(data, key);

  return index < 0 ? undefined : data[index][1];
}

/**
 * Checks if a list cache value for `key` exists.
 *
 * @private
 * @name has
 * @memberOf ListCache
 * @param {string} key The key of the entry to check.
 * @returns {boolean} Returns `true` if an entry for `key` exists, else `false`.
 */
function listCacheHas(key) {
  return assocIndexOf(this.__data__, key) > -1;
}

/**
 * Sets the list cache `key` to `value`.
 *
 * @private
 * @name set
 * @memberOf ListCache
 * @param {string} key The key of the value to set.
 * @param {*} value The value to set.
 * @returns {Object} Returns the list cache instance.
 */
function listCacheSet(key, value) {
  var data = this.__data__,
      index = assocIndexOf(data, key);

  if (index < 0) {
    data.push([key, value]);
  } else {
    data[index][1] = value;
  }
  return this;
}

// Add methods to `ListCache`.
ListCache.prototype.clear = listCacheClear;
ListCache.prototype['delete'] = listCacheDelete;
ListCache.prototype.get = listCacheGet;
ListCache.prototype.has = listCacheHas;
ListCache.prototype.set = listCacheSet;

/**
 * Creates a map cache object to store key-value pairs.
 *
 * @private
 * @constructor
 * @param {Array} [entries] The key-value pairs to cache.
 */
function MapCache(entries) {
  var index = -1,
      length = entries ? entries.length : 0;

  this.clear();
  while (++index < length) {
    var entry = entries[index];
    this.set(entry[0], entry[1]);
  }
}

/**
 * Removes all key-value entries from the map.
 *
 * @private
 * @name clear
 * @memberOf MapCache
 */
function mapCacheClear() {
  this.__data__ = {
    'hash': new Hash,
    'map': new (Map || ListCache),
    'string': new Hash
  };
}

/**
 * Removes `key` and its value from the map.
 *
 * @private
 * @name delete
 * @memberOf MapCache
 * @param {string} key The key of the value to remove.
 * @returns {boolean} Returns `true` if the entry was removed, else `false`.
 */
function mapCacheDelete(key) {
  return getMapData(this, key)['delete'](key);
}

/**
 * Gets the map value for `key`.
 *
 * @private
 * @name get
 * @memberOf MapCache
 * @param {string} key The key of the value to get.
 * @returns {*} Returns the entry value.
 */
function mapCacheGet(key) {
  return getMapData(this, key).get(key);
}

/**
 * Checks if a map value for `key` exists.
 *
 * @private
 * @name has
 * @memberOf MapCache
 * @param {string} key The key of the entry to check.
 * @returns {boolean} Returns `true` if an entry for `key` exists, else `false`.
 */
function mapCacheHas(key) {
  return getMapData(this, key).has(key);
}

/**
 * Sets the map `key` to `value`.
 *
 * @private
 * @name set
 * @memberOf MapCache
 * @param {string} key The key of the value to set.
 * @param {*} value The value to set.
 * @returns {Object} Returns the map cache instance.
 */
function mapCacheSet(key, value) {
  getMapData(this, key).set(key, value);
  return this;
}

// Add methods to `MapCache`.
MapCache.prototype.clear = mapCacheClear;
MapCache.prototype['delete'] = mapCacheDelete;
MapCache.prototype.get = mapCacheGet;
MapCache.prototype.has = mapCacheHas;
MapCache.prototype.set = mapCacheSet;

/**
 * Assigns `value` to `key` of `object` if the existing value is not equivalent
 * using [`SameValueZero`](http://ecma-international.org/ecma-262/7.0/#sec-samevaluezero)
 * for equality comparisons.
 *
 * @private
 * @param {Object} object The object to modify.
 * @param {string} key The key of the property to assign.
 * @param {*} value The value to assign.
 */
function assignValue(object, key, value) {
  var objValue = object[key];
  if (!(hasOwnProperty.call(object, key) && eq(objValue, value)) ||
      (value === undefined && !(key in object))) {
    object[key] = value;
  }
}

/**
 * Gets the index at which the `key` is found in `array` of key-value pairs.
 *
 * @private
 * @param {Array} array The array to inspect.
 * @param {*} key The key to search for.
 * @returns {number} Returns the index of the matched value, else `-1`.
 */
function assocIndexOf(array, key) {
  var length = array.length;
  while (length--) {
    if (eq(array[length][0], key)) {
      return length;
    }
  }
  return -1;
}

/**
 * The base implementation of `_.isNative` without bad shim checks.
 *
 * @private
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is a native function,
 *  else `false`.
 */
function baseIsNative(value) {
  if (!isObject$1(value) || isMasked(value)) {
    return false;
  }
  var pattern = (isFunction$1(value) || isHostObject(value)) ? reIsNative : reIsHostCtor;
  return pattern.test(toSource(value));
}

/**
 * The base implementation of `_.set`.
 *
 * @private
 * @param {Object} object The object to modify.
 * @param {Array|string} path The path of the property to set.
 * @param {*} value The value to set.
 * @param {Function} [customizer] The function to customize path creation.
 * @returns {Object} Returns `object`.
 */
function baseSet(object, path, value, customizer) {
  if (!isObject$1(object)) {
    return object;
  }
  path = isKey(path, object) ? [path] : castPath(path);

  var index = -1,
      length = path.length,
      lastIndex = length - 1,
      nested = object;

  while (nested != null && ++index < length) {
    var key = toKey(path[index]),
        newValue = value;

    if (index != lastIndex) {
      var objValue = nested[key];
      newValue = customizer ? customizer(objValue, key, nested) : undefined;
      if (newValue === undefined) {
        newValue = isObject$1(objValue)
          ? objValue
          : (isIndex(path[index + 1]) ? [] : {});
      }
    }
    assignValue(nested, key, newValue);
    nested = nested[key];
  }
  return object;
}

/**
 * The base implementation of `_.toString` which doesn't convert nullish
 * values to empty strings.
 *
 * @private
 * @param {*} value The value to process.
 * @returns {string} Returns the string.
 */
function baseToString(value) {
  // Exit early for strings to avoid a performance hit in some environments.
  if (typeof value == 'string') {
    return value;
  }
  if (isSymbol(value)) {
    return symbolToString ? symbolToString.call(value) : '';
  }
  var result = (value + '');
  return (result == '0' && (1 / value) == -INFINITY) ? '-0' : result;
}

/**
 * Casts `value` to a path array if it's not one.
 *
 * @private
 * @param {*} value The value to inspect.
 * @returns {Array} Returns the cast property path array.
 */
function castPath(value) {
  return isArray(value) ? value : stringToPath(value);
}

/**
 * Gets the data for `map`.
 *
 * @private
 * @param {Object} map The map to query.
 * @param {string} key The reference key.
 * @returns {*} Returns the map data.
 */
function getMapData(map, key) {
  var data = map.__data__;
  return isKeyable(key)
    ? data[typeof key == 'string' ? 'string' : 'hash']
    : data.map;
}

/**
 * Gets the native function at `key` of `object`.
 *
 * @private
 * @param {Object} object The object to query.
 * @param {string} key The key of the method to get.
 * @returns {*} Returns the function if it's native, else `undefined`.
 */
function getNative(object, key) {
  var value = getValue(object, key);
  return baseIsNative(value) ? value : undefined;
}

/**
 * Checks if `value` is a valid array-like index.
 *
 * @private
 * @param {*} value The value to check.
 * @param {number} [length=MAX_SAFE_INTEGER] The upper bounds of a valid index.
 * @returns {boolean} Returns `true` if `value` is a valid index, else `false`.
 */
function isIndex(value, length) {
  length = length == null ? MAX_SAFE_INTEGER : length;
  return !!length &&
    (typeof value == 'number' || reIsUint.test(value)) &&
    (value > -1 && value % 1 == 0 && value < length);
}

/**
 * Checks if `value` is a property name and not a property path.
 *
 * @private
 * @param {*} value The value to check.
 * @param {Object} [object] The object to query keys on.
 * @returns {boolean} Returns `true` if `value` is a property name, else `false`.
 */
function isKey(value, object) {
  if (isArray(value)) {
    return false;
  }
  var type = typeof value;
  if (type == 'number' || type == 'symbol' || type == 'boolean' ||
      value == null || isSymbol(value)) {
    return true;
  }
  return reIsPlainProp.test(value) || !reIsDeepProp.test(value) ||
    (object != null && value in Object(object));
}

/**
 * Checks if `value` is suitable for use as unique object key.
 *
 * @private
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is suitable, else `false`.
 */
function isKeyable(value) {
  var type = typeof value;
  return (type == 'string' || type == 'number' || type == 'symbol' || type == 'boolean')
    ? (value !== '__proto__')
    : (value === null);
}

/**
 * Checks if `func` has its source masked.
 *
 * @private
 * @param {Function} func The function to check.
 * @returns {boolean} Returns `true` if `func` is masked, else `false`.
 */
function isMasked(func) {
  return !!maskSrcKey && (maskSrcKey in func);
}

/**
 * Converts `string` to a property path array.
 *
 * @private
 * @param {string} string The string to convert.
 * @returns {Array} Returns the property path array.
 */
var stringToPath = memoize(function(string) {
  string = toString(string);

  var result = [];
  if (reLeadingDot.test(string)) {
    result.push('');
  }
  string.replace(rePropName, function(match, number, quote, string) {
    result.push(quote ? string.replace(reEscapeChar, '$1') : (number || match));
  });
  return result;
});

/**
 * Converts `value` to a string key if it's not a string or symbol.
 *
 * @private
 * @param {*} value The value to inspect.
 * @returns {string|symbol} Returns the key.
 */
function toKey(value) {
  if (typeof value == 'string' || isSymbol(value)) {
    return value;
  }
  var result = (value + '');
  return (result == '0' && (1 / value) == -INFINITY) ? '-0' : result;
}

/**
 * Converts `func` to its source code.
 *
 * @private
 * @param {Function} func The function to process.
 * @returns {string} Returns the source code.
 */
function toSource(func) {
  if (func != null) {
    try {
      return funcToString.call(func);
    } catch (e) {}
    try {
      return (func + '');
    } catch (e) {}
  }
  return '';
}

/**
 * Creates a function that memoizes the result of `func`. If `resolver` is
 * provided, it determines the cache key for storing the result based on the
 * arguments provided to the memoized function. By default, the first argument
 * provided to the memoized function is used as the map cache key. The `func`
 * is invoked with the `this` binding of the memoized function.
 *
 * **Note:** The cache is exposed as the `cache` property on the memoized
 * function. Its creation may be customized by replacing the `_.memoize.Cache`
 * constructor with one whose instances implement the
 * [`Map`](http://ecma-international.org/ecma-262/7.0/#sec-properties-of-the-map-prototype-object)
 * method interface of `delete`, `get`, `has`, and `set`.
 *
 * @static
 * @memberOf _
 * @since 0.1.0
 * @category Function
 * @param {Function} func The function to have its output memoized.
 * @param {Function} [resolver] The function to resolve the cache key.
 * @returns {Function} Returns the new memoized function.
 * @example
 *
 * var object = { 'a': 1, 'b': 2 };
 * var other = { 'c': 3, 'd': 4 };
 *
 * var values = _.memoize(_.values);
 * values(object);
 * // => [1, 2]
 *
 * values(other);
 * // => [3, 4]
 *
 * object.a = 2;
 * values(object);
 * // => [1, 2]
 *
 * // Modify the result cache.
 * values.cache.set(object, ['a', 'b']);
 * values(object);
 * // => ['a', 'b']
 *
 * // Replace `_.memoize.Cache`.
 * _.memoize.Cache = WeakMap;
 */
function memoize(func, resolver) {
  if (typeof func != 'function' || (resolver && typeof resolver != 'function')) {
    throw new TypeError(FUNC_ERROR_TEXT);
  }
  var memoized = function() {
    var args = arguments,
        key = resolver ? resolver.apply(this, args) : args[0],
        cache = memoized.cache;

    if (cache.has(key)) {
      return cache.get(key);
    }
    var result = func.apply(this, args);
    memoized.cache = cache.set(key, result);
    return result;
  };
  memoized.cache = new (memoize.Cache || MapCache);
  return memoized;
}

// Assign cache to `_.memoize`.
memoize.Cache = MapCache;

/**
 * Performs a
 * [`SameValueZero`](http://ecma-international.org/ecma-262/7.0/#sec-samevaluezero)
 * comparison between two values to determine if they are equivalent.
 *
 * @static
 * @memberOf _
 * @since 4.0.0
 * @category Lang
 * @param {*} value The value to compare.
 * @param {*} other The other value to compare.
 * @returns {boolean} Returns `true` if the values are equivalent, else `false`.
 * @example
 *
 * var object = { 'a': 1 };
 * var other = { 'a': 1 };
 *
 * _.eq(object, object);
 * // => true
 *
 * _.eq(object, other);
 * // => false
 *
 * _.eq('a', 'a');
 * // => true
 *
 * _.eq('a', Object('a'));
 * // => false
 *
 * _.eq(NaN, NaN);
 * // => true
 */
function eq(value, other) {
  return value === other || (value !== value && other !== other);
}

/**
 * Checks if `value` is classified as an `Array` object.
 *
 * @static
 * @memberOf _
 * @since 0.1.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is an array, else `false`.
 * @example
 *
 * _.isArray([1, 2, 3]);
 * // => true
 *
 * _.isArray(document.body.children);
 * // => false
 *
 * _.isArray('abc');
 * // => false
 *
 * _.isArray(_.noop);
 * // => false
 */
var isArray = Array.isArray;

/**
 * Checks if `value` is classified as a `Function` object.
 *
 * @static
 * @memberOf _
 * @since 0.1.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is a function, else `false`.
 * @example
 *
 * _.isFunction(_);
 * // => true
 *
 * _.isFunction(/abc/);
 * // => false
 */
function isFunction$1(value) {
  // The use of `Object#toString` avoids issues with the `typeof` operator
  // in Safari 8-9 which returns 'object' for typed array and other constructors.
  var tag = isObject$1(value) ? objectToString.call(value) : '';
  return tag == funcTag || tag == genTag;
}

/**
 * Checks if `value` is the
 * [language type](http://www.ecma-international.org/ecma-262/7.0/#sec-ecmascript-language-types)
 * of `Object`. (e.g. arrays, functions, objects, regexes, `new Number(0)`, and `new String('')`)
 *
 * @static
 * @memberOf _
 * @since 0.1.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is an object, else `false`.
 * @example
 *
 * _.isObject({});
 * // => true
 *
 * _.isObject([1, 2, 3]);
 * // => true
 *
 * _.isObject(_.noop);
 * // => true
 *
 * _.isObject(null);
 * // => false
 */
function isObject$1(value) {
  var type = typeof value;
  return !!value && (type == 'object' || type == 'function');
}

/**
 * Checks if `value` is object-like. A value is object-like if it's not `null`
 * and has a `typeof` result of "object".
 *
 * @static
 * @memberOf _
 * @since 4.0.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is object-like, else `false`.
 * @example
 *
 * _.isObjectLike({});
 * // => true
 *
 * _.isObjectLike([1, 2, 3]);
 * // => true
 *
 * _.isObjectLike(_.noop);
 * // => false
 *
 * _.isObjectLike(null);
 * // => false
 */
function isObjectLike(value) {
  return !!value && typeof value == 'object';
}

/**
 * Checks if `value` is classified as a `Symbol` primitive or object.
 *
 * @static
 * @memberOf _
 * @since 4.0.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is a symbol, else `false`.
 * @example
 *
 * _.isSymbol(Symbol.iterator);
 * // => true
 *
 * _.isSymbol('abc');
 * // => false
 */
function isSymbol(value) {
  return typeof value == 'symbol' ||
    (isObjectLike(value) && objectToString.call(value) == symbolTag);
}

/**
 * Converts `value` to a string. An empty string is returned for `null`
 * and `undefined` values. The sign of `-0` is preserved.
 *
 * @static
 * @memberOf _
 * @since 4.0.0
 * @category Lang
 * @param {*} value The value to process.
 * @returns {string} Returns the string.
 * @example
 *
 * _.toString(null);
 * // => ''
 *
 * _.toString(-0);
 * // => '-0'
 *
 * _.toString([1, 2, 3]);
 * // => '1,2,3'
 */
function toString(value) {
  return value == null ? '' : baseToString(value);
}

/**
 * Sets the value at `path` of `object`. If a portion of `path` doesn't exist,
 * it's created. Arrays are created for missing index properties while objects
 * are created for all other missing properties. Use `_.setWith` to customize
 * `path` creation.
 *
 * **Note:** This method mutates `object`.
 *
 * @static
 * @memberOf _
 * @since 3.7.0
 * @category Object
 * @param {Object} object The object to modify.
 * @param {Array|string} path The path of the property to set.
 * @param {*} value The value to set.
 * @returns {Object} Returns `object`.
 * @example
 *
 * var object = { 'a': [{ 'b': { 'c': 3 } }] };
 *
 * _.set(object, 'a[0].b.c', 4);
 * console.log(object.a[0].b.c);
 * // => 4
 *
 * _.set(object, ['x', '0', 'y', 'z'], 5);
 * console.log(object.x[0].y.z);
 * // => 5
 */
function set(object, path, value) {
  return object == null ? object : baseSet(object, path, value);
}

var index = set;

var index$2 = createCommonjsModule(function (module, exports) {
/**
 * lodash (Custom Build) <https://lodash.com/>
 * Build: `lodash modularize exports="npm" -o ./`
 * Copyright jQuery Foundation and other contributors <https://jquery.org/>
 * Released under MIT license <https://lodash.com/license>
 * Based on Underscore.js 1.8.3 <http://underscorejs.org/LICENSE>
 * Copyright Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
 */

/** Used as the size to enable large array optimizations. */
var LARGE_ARRAY_SIZE = 200;

/** Used to stand-in for `undefined` hash values. */
var HASH_UNDEFINED = '__lodash_hash_undefined__';

/** Used as references for various `Number` constants. */
var MAX_SAFE_INTEGER = 9007199254740991;

/** `Object#toString` result references. */
var argsTag = '[object Arguments]',
    arrayTag = '[object Array]',
    boolTag = '[object Boolean]',
    dateTag = '[object Date]',
    errorTag = '[object Error]',
    funcTag = '[object Function]',
    genTag = '[object GeneratorFunction]',
    mapTag = '[object Map]',
    numberTag = '[object Number]',
    objectTag = '[object Object]',
    promiseTag = '[object Promise]',
    regexpTag = '[object RegExp]',
    setTag = '[object Set]',
    stringTag = '[object String]',
    symbolTag = '[object Symbol]',
    weakMapTag = '[object WeakMap]';

var arrayBufferTag = '[object ArrayBuffer]',
    dataViewTag = '[object DataView]',
    float32Tag = '[object Float32Array]',
    float64Tag = '[object Float64Array]',
    int8Tag = '[object Int8Array]',
    int16Tag = '[object Int16Array]',
    int32Tag = '[object Int32Array]',
    uint8Tag = '[object Uint8Array]',
    uint8ClampedTag = '[object Uint8ClampedArray]',
    uint16Tag = '[object Uint16Array]',
    uint32Tag = '[object Uint32Array]';

/**
 * Used to match `RegExp`
 * [syntax characters](http://ecma-international.org/ecma-262/7.0/#sec-patterns).
 */
var reRegExpChar = /[\\^$.*+?()[\]{}|]/g;

/** Used to match `RegExp` flags from their coerced string values. */
var reFlags = /\w*$/;

/** Used to detect host constructors (Safari). */
var reIsHostCtor = /^\[object .+?Constructor\]$/;

/** Used to detect unsigned integer values. */
var reIsUint = /^(?:0|[1-9]\d*)$/;

/** Used to identify `toStringTag` values supported by `_.clone`. */
var cloneableTags = {};
cloneableTags[argsTag] = cloneableTags[arrayTag] =
cloneableTags[arrayBufferTag] = cloneableTags[dataViewTag] =
cloneableTags[boolTag] = cloneableTags[dateTag] =
cloneableTags[float32Tag] = cloneableTags[float64Tag] =
cloneableTags[int8Tag] = cloneableTags[int16Tag] =
cloneableTags[int32Tag] = cloneableTags[mapTag] =
cloneableTags[numberTag] = cloneableTags[objectTag] =
cloneableTags[regexpTag] = cloneableTags[setTag] =
cloneableTags[stringTag] = cloneableTags[symbolTag] =
cloneableTags[uint8Tag] = cloneableTags[uint8ClampedTag] =
cloneableTags[uint16Tag] = cloneableTags[uint32Tag] = true;
cloneableTags[errorTag] = cloneableTags[funcTag] =
cloneableTags[weakMapTag] = false;

/** Detect free variable `global` from Node.js. */
var freeGlobal = typeof commonjsGlobal == 'object' && commonjsGlobal && commonjsGlobal.Object === Object && commonjsGlobal;

/** Detect free variable `self`. */
var freeSelf = typeof self == 'object' && self && self.Object === Object && self;

/** Used as a reference to the global object. */
var root = freeGlobal || freeSelf || Function('return this')();

/** Detect free variable `exports`. */
var freeExports = 'object' == 'object' && exports && !exports.nodeType && exports;

/** Detect free variable `module`. */
var freeModule = freeExports && 'object' == 'object' && module && !module.nodeType && module;

/** Detect the popular CommonJS extension `module.exports`. */
var moduleExports = freeModule && freeModule.exports === freeExports;

/**
 * Adds the key-value `pair` to `map`.
 *
 * @private
 * @param {Object} map The map to modify.
 * @param {Array} pair The key-value pair to add.
 * @returns {Object} Returns `map`.
 */
function addMapEntry(map, pair) {
  // Don't return `map.set` because it's not chainable in IE 11.
  map.set(pair[0], pair[1]);
  return map;
}

/**
 * Adds `value` to `set`.
 *
 * @private
 * @param {Object} set The set to modify.
 * @param {*} value The value to add.
 * @returns {Object} Returns `set`.
 */
function addSetEntry(set, value) {
  // Don't return `set.add` because it's not chainable in IE 11.
  set.add(value);
  return set;
}

/**
 * A specialized version of `_.forEach` for arrays without support for
 * iteratee shorthands.
 *
 * @private
 * @param {Array} [array] The array to iterate over.
 * @param {Function} iteratee The function invoked per iteration.
 * @returns {Array} Returns `array`.
 */
function arrayEach(array, iteratee) {
  var index = -1,
      length = array ? array.length : 0;

  while (++index < length) {
    if (iteratee(array[index], index, array) === false) {
      break;
    }
  }
  return array;
}

/**
 * Appends the elements of `values` to `array`.
 *
 * @private
 * @param {Array} array The array to modify.
 * @param {Array} values The values to append.
 * @returns {Array} Returns `array`.
 */
function arrayPush(array, values) {
  var index = -1,
      length = values.length,
      offset = array.length;

  while (++index < length) {
    array[offset + index] = values[index];
  }
  return array;
}

/**
 * A specialized version of `_.reduce` for arrays without support for
 * iteratee shorthands.
 *
 * @private
 * @param {Array} [array] The array to iterate over.
 * @param {Function} iteratee The function invoked per iteration.
 * @param {*} [accumulator] The initial value.
 * @param {boolean} [initAccum] Specify using the first element of `array` as
 *  the initial value.
 * @returns {*} Returns the accumulated value.
 */
function arrayReduce(array, iteratee, accumulator, initAccum) {
  var index = -1,
      length = array ? array.length : 0;

  if (initAccum && length) {
    accumulator = array[++index];
  }
  while (++index < length) {
    accumulator = iteratee(accumulator, array[index], index, array);
  }
  return accumulator;
}

/**
 * The base implementation of `_.times` without support for iteratee shorthands
 * or max array length checks.
 *
 * @private
 * @param {number} n The number of times to invoke `iteratee`.
 * @param {Function} iteratee The function invoked per iteration.
 * @returns {Array} Returns the array of results.
 */
function baseTimes(n, iteratee) {
  var index = -1,
      result = Array(n);

  while (++index < n) {
    result[index] = iteratee(index);
  }
  return result;
}

/**
 * Gets the value at `key` of `object`.
 *
 * @private
 * @param {Object} [object] The object to query.
 * @param {string} key The key of the property to get.
 * @returns {*} Returns the property value.
 */
function getValue(object, key) {
  return object == null ? undefined : object[key];
}

/**
 * Checks if `value` is a host object in IE < 9.
 *
 * @private
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is a host object, else `false`.
 */
function isHostObject(value) {
  // Many host objects are `Object` objects that can coerce to strings
  // despite having improperly defined `toString` methods.
  var result = false;
  if (value != null && typeof value.toString != 'function') {
    try {
      result = !!(value + '');
    } catch (e) {}
  }
  return result;
}

/**
 * Converts `map` to its key-value pairs.
 *
 * @private
 * @param {Object} map The map to convert.
 * @returns {Array} Returns the key-value pairs.
 */
function mapToArray(map) {
  var index = -1,
      result = Array(map.size);

  map.forEach(function(value, key) {
    result[++index] = [key, value];
  });
  return result;
}

/**
 * Creates a unary function that invokes `func` with its argument transformed.
 *
 * @private
 * @param {Function} func The function to wrap.
 * @param {Function} transform The argument transform.
 * @returns {Function} Returns the new function.
 */
function overArg(func, transform) {
  return function(arg) {
    return func(transform(arg));
  };
}

/**
 * Converts `set` to an array of its values.
 *
 * @private
 * @param {Object} set The set to convert.
 * @returns {Array} Returns the values.
 */
function setToArray(set) {
  var index = -1,
      result = Array(set.size);

  set.forEach(function(value) {
    result[++index] = value;
  });
  return result;
}

/** Used for built-in method references. */
var arrayProto = Array.prototype,
    funcProto = Function.prototype,
    objectProto = Object.prototype;

/** Used to detect overreaching core-js shims. */
var coreJsData = root['__core-js_shared__'];

/** Used to detect methods masquerading as native. */
var maskSrcKey = (function() {
  var uid = /[^.]+$/.exec(coreJsData && coreJsData.keys && coreJsData.keys.IE_PROTO || '');
  return uid ? ('Symbol(src)_1.' + uid) : '';
}());

/** Used to resolve the decompiled source of functions. */
var funcToString = funcProto.toString;

/** Used to check objects for own properties. */
var hasOwnProperty = objectProto.hasOwnProperty;

/**
 * Used to resolve the
 * [`toStringTag`](http://ecma-international.org/ecma-262/7.0/#sec-object.prototype.tostring)
 * of values.
 */
var objectToString = objectProto.toString;

/** Used to detect if a method is native. */
var reIsNative = RegExp('^' +
  funcToString.call(hasOwnProperty).replace(reRegExpChar, '\\$&')
  .replace(/hasOwnProperty|(function).*?(?=\\\()| for .+?(?=\\\])/g, '$1.*?') + '$'
);

/** Built-in value references. */
var Buffer = moduleExports ? root.Buffer : undefined,
    Symbol = root.Symbol,
    Uint8Array = root.Uint8Array,
    getPrototype = overArg(Object.getPrototypeOf, Object),
    objectCreate = Object.create,
    propertyIsEnumerable = objectProto.propertyIsEnumerable,
    splice = arrayProto.splice;

/* Built-in method references for those with the same name as other `lodash` methods. */
var nativeGetSymbols = Object.getOwnPropertySymbols,
    nativeIsBuffer = Buffer ? Buffer.isBuffer : undefined,
    nativeKeys = overArg(Object.keys, Object);

/* Built-in method references that are verified to be native. */
var DataView = getNative(root, 'DataView'),
    Map = getNative(root, 'Map'),
    Promise = getNative(root, 'Promise'),
    Set = getNative(root, 'Set'),
    WeakMap = getNative(root, 'WeakMap'),
    nativeCreate = getNative(Object, 'create');

/** Used to detect maps, sets, and weakmaps. */
var dataViewCtorString = toSource(DataView),
    mapCtorString = toSource(Map),
    promiseCtorString = toSource(Promise),
    setCtorString = toSource(Set),
    weakMapCtorString = toSource(WeakMap);

/** Used to convert symbols to primitives and strings. */
var symbolProto = Symbol ? Symbol.prototype : undefined,
    symbolValueOf = symbolProto ? symbolProto.valueOf : undefined;

/**
 * Creates a hash object.
 *
 * @private
 * @constructor
 * @param {Array} [entries] The key-value pairs to cache.
 */
function Hash(entries) {
  var index = -1,
      length = entries ? entries.length : 0;

  this.clear();
  while (++index < length) {
    var entry = entries[index];
    this.set(entry[0], entry[1]);
  }
}

/**
 * Removes all key-value entries from the hash.
 *
 * @private
 * @name clear
 * @memberOf Hash
 */
function hashClear() {
  this.__data__ = nativeCreate ? nativeCreate(null) : {};
}

/**
 * Removes `key` and its value from the hash.
 *
 * @private
 * @name delete
 * @memberOf Hash
 * @param {Object} hash The hash to modify.
 * @param {string} key The key of the value to remove.
 * @returns {boolean} Returns `true` if the entry was removed, else `false`.
 */
function hashDelete(key) {
  return this.has(key) && delete this.__data__[key];
}

/**
 * Gets the hash value for `key`.
 *
 * @private
 * @name get
 * @memberOf Hash
 * @param {string} key The key of the value to get.
 * @returns {*} Returns the entry value.
 */
function hashGet(key) {
  var data = this.__data__;
  if (nativeCreate) {
    var result = data[key];
    return result === HASH_UNDEFINED ? undefined : result;
  }
  return hasOwnProperty.call(data, key) ? data[key] : undefined;
}

/**
 * Checks if a hash value for `key` exists.
 *
 * @private
 * @name has
 * @memberOf Hash
 * @param {string} key The key of the entry to check.
 * @returns {boolean} Returns `true` if an entry for `key` exists, else `false`.
 */
function hashHas(key) {
  var data = this.__data__;
  return nativeCreate ? data[key] !== undefined : hasOwnProperty.call(data, key);
}

/**
 * Sets the hash `key` to `value`.
 *
 * @private
 * @name set
 * @memberOf Hash
 * @param {string} key The key of the value to set.
 * @param {*} value The value to set.
 * @returns {Object} Returns the hash instance.
 */
function hashSet(key, value) {
  var data = this.__data__;
  data[key] = (nativeCreate && value === undefined) ? HASH_UNDEFINED : value;
  return this;
}

// Add methods to `Hash`.
Hash.prototype.clear = hashClear;
Hash.prototype['delete'] = hashDelete;
Hash.prototype.get = hashGet;
Hash.prototype.has = hashHas;
Hash.prototype.set = hashSet;

/**
 * Creates an list cache object.
 *
 * @private
 * @constructor
 * @param {Array} [entries] The key-value pairs to cache.
 */
function ListCache(entries) {
  var index = -1,
      length = entries ? entries.length : 0;

  this.clear();
  while (++index < length) {
    var entry = entries[index];
    this.set(entry[0], entry[1]);
  }
}

/**
 * Removes all key-value entries from the list cache.
 *
 * @private
 * @name clear
 * @memberOf ListCache
 */
function listCacheClear() {
  this.__data__ = [];
}

/**
 * Removes `key` and its value from the list cache.
 *
 * @private
 * @name delete
 * @memberOf ListCache
 * @param {string} key The key of the value to remove.
 * @returns {boolean} Returns `true` if the entry was removed, else `false`.
 */
function listCacheDelete(key) {
  var data = this.__data__,
      index = assocIndexOf(data, key);

  if (index < 0) {
    return false;
  }
  var lastIndex = data.length - 1;
  if (index == lastIndex) {
    data.pop();
  } else {
    splice.call(data, index, 1);
  }
  return true;
}

/**
 * Gets the list cache value for `key`.
 *
 * @private
 * @name get
 * @memberOf ListCache
 * @param {string} key The key of the value to get.
 * @returns {*} Returns the entry value.
 */
function listCacheGet(key) {
  var data = this.__data__,
      index = assocIndexOf(data, key);

  return index < 0 ? undefined : data[index][1];
}

/**
 * Checks if a list cache value for `key` exists.
 *
 * @private
 * @name has
 * @memberOf ListCache
 * @param {string} key The key of the entry to check.
 * @returns {boolean} Returns `true` if an entry for `key` exists, else `false`.
 */
function listCacheHas(key) {
  return assocIndexOf(this.__data__, key) > -1;
}

/**
 * Sets the list cache `key` to `value`.
 *
 * @private
 * @name set
 * @memberOf ListCache
 * @param {string} key The key of the value to set.
 * @param {*} value The value to set.
 * @returns {Object} Returns the list cache instance.
 */
function listCacheSet(key, value) {
  var data = this.__data__,
      index = assocIndexOf(data, key);

  if (index < 0) {
    data.push([key, value]);
  } else {
    data[index][1] = value;
  }
  return this;
}

// Add methods to `ListCache`.
ListCache.prototype.clear = listCacheClear;
ListCache.prototype['delete'] = listCacheDelete;
ListCache.prototype.get = listCacheGet;
ListCache.prototype.has = listCacheHas;
ListCache.prototype.set = listCacheSet;

/**
 * Creates a map cache object to store key-value pairs.
 *
 * @private
 * @constructor
 * @param {Array} [entries] The key-value pairs to cache.
 */
function MapCache(entries) {
  var index = -1,
      length = entries ? entries.length : 0;

  this.clear();
  while (++index < length) {
    var entry = entries[index];
    this.set(entry[0], entry[1]);
  }
}

/**
 * Removes all key-value entries from the map.
 *
 * @private
 * @name clear
 * @memberOf MapCache
 */
function mapCacheClear() {
  this.__data__ = {
    'hash': new Hash,
    'map': new (Map || ListCache),
    'string': new Hash
  };
}

/**
 * Removes `key` and its value from the map.
 *
 * @private
 * @name delete
 * @memberOf MapCache
 * @param {string} key The key of the value to remove.
 * @returns {boolean} Returns `true` if the entry was removed, else `false`.
 */
function mapCacheDelete(key) {
  return getMapData(this, key)['delete'](key);
}

/**
 * Gets the map value for `key`.
 *
 * @private
 * @name get
 * @memberOf MapCache
 * @param {string} key The key of the value to get.
 * @returns {*} Returns the entry value.
 */
function mapCacheGet(key) {
  return getMapData(this, key).get(key);
}

/**
 * Checks if a map value for `key` exists.
 *
 * @private
 * @name has
 * @memberOf MapCache
 * @param {string} key The key of the entry to check.
 * @returns {boolean} Returns `true` if an entry for `key` exists, else `false`.
 */
function mapCacheHas(key) {
  return getMapData(this, key).has(key);
}

/**
 * Sets the map `key` to `value`.
 *
 * @private
 * @name set
 * @memberOf MapCache
 * @param {string} key The key of the value to set.
 * @param {*} value The value to set.
 * @returns {Object} Returns the map cache instance.
 */
function mapCacheSet(key, value) {
  getMapData(this, key).set(key, value);
  return this;
}

// Add methods to `MapCache`.
MapCache.prototype.clear = mapCacheClear;
MapCache.prototype['delete'] = mapCacheDelete;
MapCache.prototype.get = mapCacheGet;
MapCache.prototype.has = mapCacheHas;
MapCache.prototype.set = mapCacheSet;

/**
 * Creates a stack cache object to store key-value pairs.
 *
 * @private
 * @constructor
 * @param {Array} [entries] The key-value pairs to cache.
 */
function Stack(entries) {
  this.__data__ = new ListCache(entries);
}

/**
 * Removes all key-value entries from the stack.
 *
 * @private
 * @name clear
 * @memberOf Stack
 */
function stackClear() {
  this.__data__ = new ListCache;
}

/**
 * Removes `key` and its value from the stack.
 *
 * @private
 * @name delete
 * @memberOf Stack
 * @param {string} key The key of the value to remove.
 * @returns {boolean} Returns `true` if the entry was removed, else `false`.
 */
function stackDelete(key) {
  return this.__data__['delete'](key);
}

/**
 * Gets the stack value for `key`.
 *
 * @private
 * @name get
 * @memberOf Stack
 * @param {string} key The key of the value to get.
 * @returns {*} Returns the entry value.
 */
function stackGet(key) {
  return this.__data__.get(key);
}

/**
 * Checks if a stack value for `key` exists.
 *
 * @private
 * @name has
 * @memberOf Stack
 * @param {string} key The key of the entry to check.
 * @returns {boolean} Returns `true` if an entry for `key` exists, else `false`.
 */
function stackHas(key) {
  return this.__data__.has(key);
}

/**
 * Sets the stack `key` to `value`.
 *
 * @private
 * @name set
 * @memberOf Stack
 * @param {string} key The key of the value to set.
 * @param {*} value The value to set.
 * @returns {Object} Returns the stack cache instance.
 */
function stackSet(key, value) {
  var cache = this.__data__;
  if (cache instanceof ListCache) {
    var pairs = cache.__data__;
    if (!Map || (pairs.length < LARGE_ARRAY_SIZE - 1)) {
      pairs.push([key, value]);
      return this;
    }
    cache = this.__data__ = new MapCache(pairs);
  }
  cache.set(key, value);
  return this;
}

// Add methods to `Stack`.
Stack.prototype.clear = stackClear;
Stack.prototype['delete'] = stackDelete;
Stack.prototype.get = stackGet;
Stack.prototype.has = stackHas;
Stack.prototype.set = stackSet;

/**
 * Creates an array of the enumerable property names of the array-like `value`.
 *
 * @private
 * @param {*} value The value to query.
 * @param {boolean} inherited Specify returning inherited property names.
 * @returns {Array} Returns the array of property names.
 */
function arrayLikeKeys(value, inherited) {
  // Safari 8.1 makes `arguments.callee` enumerable in strict mode.
  // Safari 9 makes `arguments.length` enumerable in strict mode.
  var result = (isArray(value) || isArguments(value))
    ? baseTimes(value.length, String)
    : [];

  var length = result.length,
      skipIndexes = !!length;

  for (var key in value) {
    if ((inherited || hasOwnProperty.call(value, key)) &&
        !(skipIndexes && (key == 'length' || isIndex(key, length)))) {
      result.push(key);
    }
  }
  return result;
}

/**
 * Assigns `value` to `key` of `object` if the existing value is not equivalent
 * using [`SameValueZero`](http://ecma-international.org/ecma-262/7.0/#sec-samevaluezero)
 * for equality comparisons.
 *
 * @private
 * @param {Object} object The object to modify.
 * @param {string} key The key of the property to assign.
 * @param {*} value The value to assign.
 */
function assignValue(object, key, value) {
  var objValue = object[key];
  if (!(hasOwnProperty.call(object, key) && eq(objValue, value)) ||
      (value === undefined && !(key in object))) {
    object[key] = value;
  }
}

/**
 * Gets the index at which the `key` is found in `array` of key-value pairs.
 *
 * @private
 * @param {Array} array The array to inspect.
 * @param {*} key The key to search for.
 * @returns {number} Returns the index of the matched value, else `-1`.
 */
function assocIndexOf(array, key) {
  var length = array.length;
  while (length--) {
    if (eq(array[length][0], key)) {
      return length;
    }
  }
  return -1;
}

/**
 * The base implementation of `_.assign` without support for multiple sources
 * or `customizer` functions.
 *
 * @private
 * @param {Object} object The destination object.
 * @param {Object} source The source object.
 * @returns {Object} Returns `object`.
 */
function baseAssign(object, source) {
  return object && copyObject(source, keys(source), object);
}

/**
 * The base implementation of `_.clone` and `_.cloneDeep` which tracks
 * traversed objects.
 *
 * @private
 * @param {*} value The value to clone.
 * @param {boolean} [isDeep] Specify a deep clone.
 * @param {boolean} [isFull] Specify a clone including symbols.
 * @param {Function} [customizer] The function to customize cloning.
 * @param {string} [key] The key of `value`.
 * @param {Object} [object] The parent object of `value`.
 * @param {Object} [stack] Tracks traversed objects and their clone counterparts.
 * @returns {*} Returns the cloned value.
 */
function baseClone(value, isDeep, isFull, customizer, key, object, stack) {
  var result;
  if (customizer) {
    result = object ? customizer(value, key, object, stack) : customizer(value);
  }
  if (result !== undefined) {
    return result;
  }
  if (!isObject(value)) {
    return value;
  }
  var isArr = isArray(value);
  if (isArr) {
    result = initCloneArray(value);
    if (!isDeep) {
      return copyArray(value, result);
    }
  } else {
    var tag = getTag(value),
        isFunc = tag == funcTag || tag == genTag;

    if (isBuffer(value)) {
      return cloneBuffer(value, isDeep);
    }
    if (tag == objectTag || tag == argsTag || (isFunc && !object)) {
      if (isHostObject(value)) {
        return object ? value : {};
      }
      result = initCloneObject(isFunc ? {} : value);
      if (!isDeep) {
        return copySymbols(value, baseAssign(result, value));
      }
    } else {
      if (!cloneableTags[tag]) {
        return object ? value : {};
      }
      result = initCloneByTag(value, tag, baseClone, isDeep);
    }
  }
  // Check for circular references and return its corresponding clone.
  stack || (stack = new Stack);
  var stacked = stack.get(value);
  if (stacked) {
    return stacked;
  }
  stack.set(value, result);

  if (!isArr) {
    var props = isFull ? getAllKeys(value) : keys(value);
  }
  arrayEach(props || value, function(subValue, key) {
    if (props) {
      key = subValue;
      subValue = value[key];
    }
    // Recursively populate clone (susceptible to call stack limits).
    assignValue(result, key, baseClone(subValue, isDeep, isFull, customizer, key, value, stack));
  });
  return result;
}

/**
 * The base implementation of `_.create` without support for assigning
 * properties to the created object.
 *
 * @private
 * @param {Object} prototype The object to inherit from.
 * @returns {Object} Returns the new object.
 */
function baseCreate(proto) {
  return isObject(proto) ? objectCreate(proto) : {};
}

/**
 * The base implementation of `getAllKeys` and `getAllKeysIn` which uses
 * `keysFunc` and `symbolsFunc` to get the enumerable property names and
 * symbols of `object`.
 *
 * @private
 * @param {Object} object The object to query.
 * @param {Function} keysFunc The function to get the keys of `object`.
 * @param {Function} symbolsFunc The function to get the symbols of `object`.
 * @returns {Array} Returns the array of property names and symbols.
 */
function baseGetAllKeys(object, keysFunc, symbolsFunc) {
  var result = keysFunc(object);
  return isArray(object) ? result : arrayPush(result, symbolsFunc(object));
}

/**
 * The base implementation of `getTag`.
 *
 * @private
 * @param {*} value The value to query.
 * @returns {string} Returns the `toStringTag`.
 */
function baseGetTag(value) {
  return objectToString.call(value);
}

/**
 * The base implementation of `_.isNative` without bad shim checks.
 *
 * @private
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is a native function,
 *  else `false`.
 */
function baseIsNative(value) {
  if (!isObject(value) || isMasked(value)) {
    return false;
  }
  var pattern = (isFunction(value) || isHostObject(value)) ? reIsNative : reIsHostCtor;
  return pattern.test(toSource(value));
}

/**
 * The base implementation of `_.keys` which doesn't treat sparse arrays as dense.
 *
 * @private
 * @param {Object} object The object to query.
 * @returns {Array} Returns the array of property names.
 */
function baseKeys(object) {
  if (!isPrototype(object)) {
    return nativeKeys(object);
  }
  var result = [];
  for (var key in Object(object)) {
    if (hasOwnProperty.call(object, key) && key != 'constructor') {
      result.push(key);
    }
  }
  return result;
}

/**
 * Creates a clone of  `buffer`.
 *
 * @private
 * @param {Buffer} buffer The buffer to clone.
 * @param {boolean} [isDeep] Specify a deep clone.
 * @returns {Buffer} Returns the cloned buffer.
 */
function cloneBuffer(buffer, isDeep) {
  if (isDeep) {
    return buffer.slice();
  }
  var result = new buffer.constructor(buffer.length);
  buffer.copy(result);
  return result;
}

/**
 * Creates a clone of `arrayBuffer`.
 *
 * @private
 * @param {ArrayBuffer} arrayBuffer The array buffer to clone.
 * @returns {ArrayBuffer} Returns the cloned array buffer.
 */
function cloneArrayBuffer(arrayBuffer) {
  var result = new arrayBuffer.constructor(arrayBuffer.byteLength);
  new Uint8Array(result).set(new Uint8Array(arrayBuffer));
  return result;
}

/**
 * Creates a clone of `dataView`.
 *
 * @private
 * @param {Object} dataView The data view to clone.
 * @param {boolean} [isDeep] Specify a deep clone.
 * @returns {Object} Returns the cloned data view.
 */
function cloneDataView(dataView, isDeep) {
  var buffer = isDeep ? cloneArrayBuffer(dataView.buffer) : dataView.buffer;
  return new dataView.constructor(buffer, dataView.byteOffset, dataView.byteLength);
}

/**
 * Creates a clone of `map`.
 *
 * @private
 * @param {Object} map The map to clone.
 * @param {Function} cloneFunc The function to clone values.
 * @param {boolean} [isDeep] Specify a deep clone.
 * @returns {Object} Returns the cloned map.
 */
function cloneMap(map, isDeep, cloneFunc) {
  var array = isDeep ? cloneFunc(mapToArray(map), true) : mapToArray(map);
  return arrayReduce(array, addMapEntry, new map.constructor);
}

/**
 * Creates a clone of `regexp`.
 *
 * @private
 * @param {Object} regexp The regexp to clone.
 * @returns {Object} Returns the cloned regexp.
 */
function cloneRegExp(regexp) {
  var result = new regexp.constructor(regexp.source, reFlags.exec(regexp));
  result.lastIndex = regexp.lastIndex;
  return result;
}

/**
 * Creates a clone of `set`.
 *
 * @private
 * @param {Object} set The set to clone.
 * @param {Function} cloneFunc The function to clone values.
 * @param {boolean} [isDeep] Specify a deep clone.
 * @returns {Object} Returns the cloned set.
 */
function cloneSet(set, isDeep, cloneFunc) {
  var array = isDeep ? cloneFunc(setToArray(set), true) : setToArray(set);
  return arrayReduce(array, addSetEntry, new set.constructor);
}

/**
 * Creates a clone of the `symbol` object.
 *
 * @private
 * @param {Object} symbol The symbol object to clone.
 * @returns {Object} Returns the cloned symbol object.
 */
function cloneSymbol(symbol) {
  return symbolValueOf ? Object(symbolValueOf.call(symbol)) : {};
}

/**
 * Creates a clone of `typedArray`.
 *
 * @private
 * @param {Object} typedArray The typed array to clone.
 * @param {boolean} [isDeep] Specify a deep clone.
 * @returns {Object} Returns the cloned typed array.
 */
function cloneTypedArray(typedArray, isDeep) {
  var buffer = isDeep ? cloneArrayBuffer(typedArray.buffer) : typedArray.buffer;
  return new typedArray.constructor(buffer, typedArray.byteOffset, typedArray.length);
}

/**
 * Copies the values of `source` to `array`.
 *
 * @private
 * @param {Array} source The array to copy values from.
 * @param {Array} [array=[]] The array to copy values to.
 * @returns {Array} Returns `array`.
 */
function copyArray(source, array) {
  var index = -1,
      length = source.length;

  array || (array = Array(length));
  while (++index < length) {
    array[index] = source[index];
  }
  return array;
}

/**
 * Copies properties of `source` to `object`.
 *
 * @private
 * @param {Object} source The object to copy properties from.
 * @param {Array} props The property identifiers to copy.
 * @param {Object} [object={}] The object to copy properties to.
 * @param {Function} [customizer] The function to customize copied values.
 * @returns {Object} Returns `object`.
 */
function copyObject(source, props, object, customizer) {
  object || (object = {});

  var index = -1,
      length = props.length;

  while (++index < length) {
    var key = props[index];

    var newValue = customizer
      ? customizer(object[key], source[key], key, object, source)
      : undefined;

    assignValue(object, key, newValue === undefined ? source[key] : newValue);
  }
  return object;
}

/**
 * Copies own symbol properties of `source` to `object`.
 *
 * @private
 * @param {Object} source The object to copy symbols from.
 * @param {Object} [object={}] The object to copy symbols to.
 * @returns {Object} Returns `object`.
 */
function copySymbols(source, object) {
  return copyObject(source, getSymbols(source), object);
}

/**
 * Creates an array of own enumerable property names and symbols of `object`.
 *
 * @private
 * @param {Object} object The object to query.
 * @returns {Array} Returns the array of property names and symbols.
 */
function getAllKeys(object) {
  return baseGetAllKeys(object, keys, getSymbols);
}

/**
 * Gets the data for `map`.
 *
 * @private
 * @param {Object} map The map to query.
 * @param {string} key The reference key.
 * @returns {*} Returns the map data.
 */
function getMapData(map, key) {
  var data = map.__data__;
  return isKeyable(key)
    ? data[typeof key == 'string' ? 'string' : 'hash']
    : data.map;
}

/**
 * Gets the native function at `key` of `object`.
 *
 * @private
 * @param {Object} object The object to query.
 * @param {string} key The key of the method to get.
 * @returns {*} Returns the function if it's native, else `undefined`.
 */
function getNative(object, key) {
  var value = getValue(object, key);
  return baseIsNative(value) ? value : undefined;
}

/**
 * Creates an array of the own enumerable symbol properties of `object`.
 *
 * @private
 * @param {Object} object The object to query.
 * @returns {Array} Returns the array of symbols.
 */
var getSymbols = nativeGetSymbols ? overArg(nativeGetSymbols, Object) : stubArray;

/**
 * Gets the `toStringTag` of `value`.
 *
 * @private
 * @param {*} value The value to query.
 * @returns {string} Returns the `toStringTag`.
 */
var getTag = baseGetTag;

// Fallback for data views, maps, sets, and weak maps in IE 11,
// for data views in Edge < 14, and promises in Node.js.
if ((DataView && getTag(new DataView(new ArrayBuffer(1))) != dataViewTag) ||
    (Map && getTag(new Map) != mapTag) ||
    (Promise && getTag(Promise.resolve()) != promiseTag) ||
    (Set && getTag(new Set) != setTag) ||
    (WeakMap && getTag(new WeakMap) != weakMapTag)) {
  getTag = function(value) {
    var result = objectToString.call(value),
        Ctor = result == objectTag ? value.constructor : undefined,
        ctorString = Ctor ? toSource(Ctor) : undefined;

    if (ctorString) {
      switch (ctorString) {
        case dataViewCtorString: return dataViewTag;
        case mapCtorString: return mapTag;
        case promiseCtorString: return promiseTag;
        case setCtorString: return setTag;
        case weakMapCtorString: return weakMapTag;
      }
    }
    return result;
  };
}

/**
 * Initializes an array clone.
 *
 * @private
 * @param {Array} array The array to clone.
 * @returns {Array} Returns the initialized clone.
 */
function initCloneArray(array) {
  var length = array.length,
      result = array.constructor(length);

  // Add properties assigned by `RegExp#exec`.
  if (length && typeof array[0] == 'string' && hasOwnProperty.call(array, 'index')) {
    result.index = array.index;
    result.input = array.input;
  }
  return result;
}

/**
 * Initializes an object clone.
 *
 * @private
 * @param {Object} object The object to clone.
 * @returns {Object} Returns the initialized clone.
 */
function initCloneObject(object) {
  return (typeof object.constructor == 'function' && !isPrototype(object))
    ? baseCreate(getPrototype(object))
    : {};
}

/**
 * Initializes an object clone based on its `toStringTag`.
 *
 * **Note:** This function only supports cloning values with tags of
 * `Boolean`, `Date`, `Error`, `Number`, `RegExp`, or `String`.
 *
 * @private
 * @param {Object} object The object to clone.
 * @param {string} tag The `toStringTag` of the object to clone.
 * @param {Function} cloneFunc The function to clone values.
 * @param {boolean} [isDeep] Specify a deep clone.
 * @returns {Object} Returns the initialized clone.
 */
function initCloneByTag(object, tag, cloneFunc, isDeep) {
  var Ctor = object.constructor;
  switch (tag) {
    case arrayBufferTag:
      return cloneArrayBuffer(object);

    case boolTag:
    case dateTag:
      return new Ctor(+object);

    case dataViewTag:
      return cloneDataView(object, isDeep);

    case float32Tag: case float64Tag:
    case int8Tag: case int16Tag: case int32Tag:
    case uint8Tag: case uint8ClampedTag: case uint16Tag: case uint32Tag:
      return cloneTypedArray(object, isDeep);

    case mapTag:
      return cloneMap(object, isDeep, cloneFunc);

    case numberTag:
    case stringTag:
      return new Ctor(object);

    case regexpTag:
      return cloneRegExp(object);

    case setTag:
      return cloneSet(object, isDeep, cloneFunc);

    case symbolTag:
      return cloneSymbol(object);
  }
}

/**
 * Checks if `value` is a valid array-like index.
 *
 * @private
 * @param {*} value The value to check.
 * @param {number} [length=MAX_SAFE_INTEGER] The upper bounds of a valid index.
 * @returns {boolean} Returns `true` if `value` is a valid index, else `false`.
 */
function isIndex(value, length) {
  length = length == null ? MAX_SAFE_INTEGER : length;
  return !!length &&
    (typeof value == 'number' || reIsUint.test(value)) &&
    (value > -1 && value % 1 == 0 && value < length);
}

/**
 * Checks if `value` is suitable for use as unique object key.
 *
 * @private
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is suitable, else `false`.
 */
function isKeyable(value) {
  var type = typeof value;
  return (type == 'string' || type == 'number' || type == 'symbol' || type == 'boolean')
    ? (value !== '__proto__')
    : (value === null);
}

/**
 * Checks if `func` has its source masked.
 *
 * @private
 * @param {Function} func The function to check.
 * @returns {boolean} Returns `true` if `func` is masked, else `false`.
 */
function isMasked(func) {
  return !!maskSrcKey && (maskSrcKey in func);
}

/**
 * Checks if `value` is likely a prototype object.
 *
 * @private
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is a prototype, else `false`.
 */
function isPrototype(value) {
  var Ctor = value && value.constructor,
      proto = (typeof Ctor == 'function' && Ctor.prototype) || objectProto;

  return value === proto;
}

/**
 * Converts `func` to its source code.
 *
 * @private
 * @param {Function} func The function to process.
 * @returns {string} Returns the source code.
 */
function toSource(func) {
  if (func != null) {
    try {
      return funcToString.call(func);
    } catch (e) {}
    try {
      return (func + '');
    } catch (e) {}
  }
  return '';
}

/**
 * This method is like `_.clone` except that it recursively clones `value`.
 *
 * @static
 * @memberOf _
 * @since 1.0.0
 * @category Lang
 * @param {*} value The value to recursively clone.
 * @returns {*} Returns the deep cloned value.
 * @see _.clone
 * @example
 *
 * var objects = [{ 'a': 1 }, { 'b': 2 }];
 *
 * var deep = _.cloneDeep(objects);
 * console.log(deep[0] === objects[0]);
 * // => false
 */
function cloneDeep(value) {
  return baseClone(value, true, true);
}

/**
 * Performs a
 * [`SameValueZero`](http://ecma-international.org/ecma-262/7.0/#sec-samevaluezero)
 * comparison between two values to determine if they are equivalent.
 *
 * @static
 * @memberOf _
 * @since 4.0.0
 * @category Lang
 * @param {*} value The value to compare.
 * @param {*} other The other value to compare.
 * @returns {boolean} Returns `true` if the values are equivalent, else `false`.
 * @example
 *
 * var object = { 'a': 1 };
 * var other = { 'a': 1 };
 *
 * _.eq(object, object);
 * // => true
 *
 * _.eq(object, other);
 * // => false
 *
 * _.eq('a', 'a');
 * // => true
 *
 * _.eq('a', Object('a'));
 * // => false
 *
 * _.eq(NaN, NaN);
 * // => true
 */
function eq(value, other) {
  return value === other || (value !== value && other !== other);
}

/**
 * Checks if `value` is likely an `arguments` object.
 *
 * @static
 * @memberOf _
 * @since 0.1.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is an `arguments` object,
 *  else `false`.
 * @example
 *
 * _.isArguments(function() { return arguments; }());
 * // => true
 *
 * _.isArguments([1, 2, 3]);
 * // => false
 */
function isArguments(value) {
  // Safari 8.1 makes `arguments.callee` enumerable in strict mode.
  return isArrayLikeObject(value) && hasOwnProperty.call(value, 'callee') &&
    (!propertyIsEnumerable.call(value, 'callee') || objectToString.call(value) == argsTag);
}

/**
 * Checks if `value` is classified as an `Array` object.
 *
 * @static
 * @memberOf _
 * @since 0.1.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is an array, else `false`.
 * @example
 *
 * _.isArray([1, 2, 3]);
 * // => true
 *
 * _.isArray(document.body.children);
 * // => false
 *
 * _.isArray('abc');
 * // => false
 *
 * _.isArray(_.noop);
 * // => false
 */
var isArray = Array.isArray;

/**
 * Checks if `value` is array-like. A value is considered array-like if it's
 * not a function and has a `value.length` that's an integer greater than or
 * equal to `0` and less than or equal to `Number.MAX_SAFE_INTEGER`.
 *
 * @static
 * @memberOf _
 * @since 4.0.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is array-like, else `false`.
 * @example
 *
 * _.isArrayLike([1, 2, 3]);
 * // => true
 *
 * _.isArrayLike(document.body.children);
 * // => true
 *
 * _.isArrayLike('abc');
 * // => true
 *
 * _.isArrayLike(_.noop);
 * // => false
 */
function isArrayLike(value) {
  return value != null && isLength(value.length) && !isFunction(value);
}

/**
 * This method is like `_.isArrayLike` except that it also checks if `value`
 * is an object.
 *
 * @static
 * @memberOf _
 * @since 4.0.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is an array-like object,
 *  else `false`.
 * @example
 *
 * _.isArrayLikeObject([1, 2, 3]);
 * // => true
 *
 * _.isArrayLikeObject(document.body.children);
 * // => true
 *
 * _.isArrayLikeObject('abc');
 * // => false
 *
 * _.isArrayLikeObject(_.noop);
 * // => false
 */
function isArrayLikeObject(value) {
  return isObjectLike(value) && isArrayLike(value);
}

/**
 * Checks if `value` is a buffer.
 *
 * @static
 * @memberOf _
 * @since 4.3.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is a buffer, else `false`.
 * @example
 *
 * _.isBuffer(new Buffer(2));
 * // => true
 *
 * _.isBuffer(new Uint8Array(2));
 * // => false
 */
var isBuffer = nativeIsBuffer || stubFalse;

/**
 * Checks if `value` is classified as a `Function` object.
 *
 * @static
 * @memberOf _
 * @since 0.1.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is a function, else `false`.
 * @example
 *
 * _.isFunction(_);
 * // => true
 *
 * _.isFunction(/abc/);
 * // => false
 */
function isFunction(value) {
  // The use of `Object#toString` avoids issues with the `typeof` operator
  // in Safari 8-9 which returns 'object' for typed array and other constructors.
  var tag = isObject(value) ? objectToString.call(value) : '';
  return tag == funcTag || tag == genTag;
}

/**
 * Checks if `value` is a valid array-like length.
 *
 * **Note:** This method is loosely based on
 * [`ToLength`](http://ecma-international.org/ecma-262/7.0/#sec-tolength).
 *
 * @static
 * @memberOf _
 * @since 4.0.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is a valid length, else `false`.
 * @example
 *
 * _.isLength(3);
 * // => true
 *
 * _.isLength(Number.MIN_VALUE);
 * // => false
 *
 * _.isLength(Infinity);
 * // => false
 *
 * _.isLength('3');
 * // => false
 */
function isLength(value) {
  return typeof value == 'number' &&
    value > -1 && value % 1 == 0 && value <= MAX_SAFE_INTEGER;
}

/**
 * Checks if `value` is the
 * [language type](http://www.ecma-international.org/ecma-262/7.0/#sec-ecmascript-language-types)
 * of `Object`. (e.g. arrays, functions, objects, regexes, `new Number(0)`, and `new String('')`)
 *
 * @static
 * @memberOf _
 * @since 0.1.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is an object, else `false`.
 * @example
 *
 * _.isObject({});
 * // => true
 *
 * _.isObject([1, 2, 3]);
 * // => true
 *
 * _.isObject(_.noop);
 * // => true
 *
 * _.isObject(null);
 * // => false
 */
function isObject(value) {
  var type = typeof value;
  return !!value && (type == 'object' || type == 'function');
}

/**
 * Checks if `value` is object-like. A value is object-like if it's not `null`
 * and has a `typeof` result of "object".
 *
 * @static
 * @memberOf _
 * @since 4.0.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is object-like, else `false`.
 * @example
 *
 * _.isObjectLike({});
 * // => true
 *
 * _.isObjectLike([1, 2, 3]);
 * // => true
 *
 * _.isObjectLike(_.noop);
 * // => false
 *
 * _.isObjectLike(null);
 * // => false
 */
function isObjectLike(value) {
  return !!value && typeof value == 'object';
}

/**
 * Creates an array of the own enumerable property names of `object`.
 *
 * **Note:** Non-object values are coerced to objects. See the
 * [ES spec](http://ecma-international.org/ecma-262/7.0/#sec-object.keys)
 * for more details.
 *
 * @static
 * @since 0.1.0
 * @memberOf _
 * @category Object
 * @param {Object} object The object to query.
 * @returns {Array} Returns the array of property names.
 * @example
 *
 * function Foo() {
 *   this.a = 1;
 *   this.b = 2;
 * }
 *
 * Foo.prototype.c = 3;
 *
 * _.keys(new Foo);
 * // => ['a', 'b'] (iteration order is not guaranteed)
 *
 * _.keys('hi');
 * // => ['0', '1']
 */
function keys(object) {
  return isArrayLike(object) ? arrayLikeKeys(object) : baseKeys(object);
}

/**
 * This method returns a new empty array.
 *
 * @static
 * @memberOf _
 * @since 4.13.0
 * @category Util
 * @returns {Array} Returns the new empty array.
 * @example
 *
 * var arrays = _.times(2, _.stubArray);
 *
 * console.log(arrays);
 * // => [[], []]
 *
 * console.log(arrays[0] === arrays[1]);
 * // => false
 */
function stubArray() {
  return [];
}

/**
 * This method returns `false`.
 *
 * @static
 * @memberOf _
 * @since 4.13.0
 * @category Util
 * @returns {boolean} Returns `false`.
 * @example
 *
 * _.times(2, _.stubFalse);
 * // => [false, false]
 */
function stubFalse() {
  return false;
}

module.exports = cloneDeep;
});

/* eslint-disable no-use-before-define */

/**
 * @desc If value is `null` or `undefined`,
 *       then we should pass empty string as a DOM value.
 *
 */
function normalizeValueForDom(value) {
  return isDefined(value) && !isNull(value) ? value : '';
}

/**
 * @desc If key path consists of single attribute,
 *       user can pass it as a string.
 *       This function normalizes it to Array.
 *
 */
function normalizeKeyPath(keyPath) {
  return Array.isArray(keyPath) ? keyPath : [keyPath];
}

/**
 * @desc Retrives value of property at provided key path.
 *       Data container can be vanilla JS Object,
 *       as well as Immutable structure (e.g. Map).
 *
 */
function getProp(container, // eslint-disable-line flowtype/no-weak-types
keyPath) {
  var normalizedKeyPath = normalizeKeyPath(keyPath);

  return container.getIn && isFunction(container.getIn) ? container.getIn(normalizedKeyPath) : normalizedKeyPath.reduce(function (data, key) {
    return data[key];
  }, container);
}

/**
 * @desc Sets value of property at provided key path.
 *       Data container can be vanilla JS Object,
 *       as well as Immutable structure (e.g. Map).
 *
 */
function setProp(container, // eslint-disable-line flowtype/no-weak-types
keyPath, value) {
  var normalizedKeyPath = normalizeKeyPath(keyPath);

  return container.setIn && isFunction(container.setIn) ? container.setIn(normalizedKeyPath, value) : index(index$2(container), keyPath, value);
}

function isDefined(subject) {
  return typeof subject !== 'undefined';
}

function isNull(subject) {
  return subject === null;
}

function isString(subject) {
  return typeof subject === 'string';
}

function isNumber(subject) {
  return typeof subject === 'number';
}

function isObject(subject) {
  return typeof subject === 'object';
}

function isPlainObject(subject) {
  return isObject(subject) && Object.prototype.toString.call(subject) === '[object Object]';
}

function isFunction(subject) {
  return typeof subject === 'function';
}

// eslint-disable-next-line flowtype/no-weak-types
function isPromise(subject) {
  return !!subject && (isObject(subject) || isFunction(subject)) && isFunction(subject.then);
}

function isChangeEvent(event) {
  return event.type === 'change';
}

function isBlurEvent(event) {
  return event.type === 'blur';
}

var classCallCheck = function (instance, Constructor) {
  if (!(instance instanceof Constructor)) {
    throw new TypeError("Cannot call a class as a function");
  }
};









var _extends = Object.assign || function (target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i];

    for (var key in source) {
      if (Object.prototype.hasOwnProperty.call(source, key)) {
        target[key] = source[key];
      }
    }
  }

  return target;
};



var inherits = function (subClass, superClass) {
  if (typeof superClass !== "function" && superClass !== null) {
    throw new TypeError("Super expression must either be null or a function, not " + typeof superClass);
  }

  subClass.prototype = Object.create(superClass && superClass.prototype, {
    constructor: {
      value: subClass,
      enumerable: false,
      writable: true,
      configurable: true
    }
  });
  if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass;
};











var possibleConstructorReturn = function (self, call) {
  if (!self) {
    throw new ReferenceError("this hasn't been initialised - super() hasn't been called");
  }

  return call && (typeof call === "object" || typeof call === "function") ? call : self;
};

/**
 * @desc It is possible to return validation result as simple boolean,
 *       as well as Object w/ additional data (i.e. `message` & `status`).
 *       This function normalizes validation results to Object shape.
 *
 */
function normalizeValidationResults(results, layerId, fieldId) {
  if (results === true) {
    return { valid: true };
  } else if (results === false) {
    return { valid: false };
  }

  if (!isDefined(results.valid)) {
    throw new Error(buildErrorMessage({
      layerId: layerId,
      fieldId: fieldId,
      message: ["Can't find `valid` key in the results object from validator.", "Make sure it's defined and its value is Boolean"]
    }));
  }

  var normalizedResults = {
    valid: results.valid
  };

  if (results.status) normalizedResults.status = results.status;
  if (results.message) normalizedResults.message = results.message;

  return normalizedResults;
}

/**
 * @desc Applies sync validator and returns normalized validation results.
 *
 */
function applySyncValidation(validator, value, data, layerId, fieldId) {
  if (!validator) return { valid: true };

  if (validator && !isFunction(validator)) {
    throw new Error(buildErrorMessage({
      layerId: layerId,
      fieldId: fieldId,
      message: 'Validator must be a function.'
    }));
  }

  var results = validator(value, data);

  return normalizeValidationResults(results, layerId, fieldId);
}

/**
 * @desc Calculates default success & failure statuses.
 *
 */
// $FlowIssue: https://github.com/facebook/flow/issues/2977
function getDefaultStatuses() {
  var statuses = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

  return {
    success: statuses.success || DefaultStatus.SUCCESS,
    failure: statuses.failure || DefaultStatus.FAILURE
  };
}

/**
 * @desc Builds Object w/ validation results from synchronous validation.
 *       This results will be emitted to the view component.
 *
 */
function buildCompleteSyncValidationResults(field, value, data, statuses, layerId) {
  var defaultStatuses = getDefaultStatuses(statuses);
  var validationState = applySyncValidation(field.validate, value, data, layerId, field.id);

  if (!validationState.valid) {
    validationState.status = validationState.status || defaultStatuses.failure;
  } else if (value) {
    validationState.status = validationState.status || defaultStatuses.success;
  }

  return validationState;
}

/**
 * @desc Builds Object w/ validation results from asynchronous validation.
 *       This results will be emitted to the view component.
 *
 */
function buildCompleteAsyncValidationResults(fieldId, results, statuses, layerId) {
  var defaultStatuses = getDefaultStatuses(statuses);
  var validationState = _extends({}, normalizeValidationResults(results, layerId, fieldId), {
    isAsync: true
  });

  if (!validationState.valid) {
    validationState.status = validationState.status || defaultStatuses.failure;
  } else {
    validationState.status = validationState.status || defaultStatuses.success;
  }

  // $FlowIssue: exact type + destructuring: https://github.com/facebook/flow/issues/2405
  return validationState;
}

/**
 * @desc Builds Object w/ intermediate async validation results.
 *       These results will be emitted to the view component,
 *       when either debounced or non-debounced async validator
 *       is triggered and we're waiting for remote validation to be resolved.
 *
 */
function buildIntermediateAsyncValidationResults() {
  return {
    valid: null,
    status: null,
    message: null,
    isProcessing: true
  };
}

/**
 * @desc Builds Object w/ empty validation results.
 *       These results will be emitted to the view component,
 *       but it means that Validation Layer doesn't emit
 *       any results yet according to strategy.
 *
 */
function buildEmptyValidationResults() {
  return {
    valid: null,
    status: null,
    message: null
  };
}

/**
 * @desc Debounce async validation calls.
 *       Based on lodash.debounce.
 *
 */

/* eslint-disable no-use-before-define, consistent-return */

function debounce(async) {
  var wait = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 0;

  var lastContext = void 0;
  var lastValue = void 0;
  var lastFieldId = void 0;
  var lastStateContainer = void 0;
  var lastCallTime = void 0;
  var timerId = void 0;

  function shouldInvoke(time) {
    var timeSinceLastCall = time - lastCallTime;

    return lastCallTime === undefined || timeSinceLastCall >= wait || timeSinceLastCall < 0;
  }

  function initAsync() {
    // Start the timer for the invoke async
    timerId = setTimeout(timerExpired, wait);
    // Report initialized
    return { status: DebounceStatus.INITIALIZED };
  }

  function remainingWait(time) {
    var timeSinceLastCall = time - lastCallTime;
    return wait - timeSinceLastCall;
  }

  function timerExpired() {
    var time = Date.now();

    if (shouldInvoke(time)) {
      return invokeAsync(time);
    }

    timerId = setTimeout(timerExpired, remainingWait(time));
  }

  function invokeAsync() {
    timerId = undefined;

    if (lastValue) {
      return triggerRequest();
    }
  }

  function triggerRequest() {
    var context = lastContext;
    var fieldId = lastFieldId;
    var value = lastValue;
    var stateContainer = lastStateContainer;

    lastContext = lastFieldId = lastValue = lastStateContainer = undefined;

    if (!fieldId || !stateContainer) throw new Error();

    // Notify layer about triggered request
    return stateContainer.notifyAsync(fieldId, value, async.call(context, value));
  }

  return function debounced(fieldId, value, stateContainer) {
    var time = Date.now();
    var isInvoking = shouldInvoke(time);

    lastCallTime = time;
    lastContext = this;

    lastFieldId = fieldId;
    lastValue = value;
    lastStateContainer = stateContainer;

    // Initialize if it's a first ivokation
    if (isInvoking && timerId === undefined) {
      return initAsync(lastCallTime);
    }

    // Start timer if it's not done yet
    if (timerId === undefined) {
      timerId = setTimeout(timerExpired, wait);
    }

    // Report debounced
    return { status: DebounceStatus.DEBOUNCED };
  };
}

/**
 * @desc Builds internal identificator of the field,
 *       e.g. `user.email`
 *
 */
function buildFieldId(keyPath) {
  return keyPath.join(Constant.FIELD_ID_DELIMITER);
}

/**
 * @desc Parses internal identificator of the field,
 *       and returns name of the attribute and key path to its value.
 *       e.g. `email` & `['user', 'email']`
 *
 */
function parseFieldId(fieldId) {
  var keyPath = fieldId.split(Constant.FIELD_ID_DELIMITER);
  return { keyPath: keyPath, attr: keyPath[keyPath.length - 1] };
}

/**
 * @desc Builds internal identificator of the field from a key path.
 *
 */
function buildFieldIdFromUserKeyPath(userKeyPath) {
  var normalizedKeyPath = normalizeKeyPath(userKeyPath);
  return buildFieldId(normalizedKeyPath);
}

/**
 * @desc Builds DOM id attribute of the field,
 *       including layerId to namespace it,
 *       e.g. `loginForm___user___email`
 *
 */
function buildFieldDomId(layerId, keyPath) {
  return [layerId].concat(keyPath).join(Constant.FIELD_DOM_ID_DELIMITER);
}

/**
 * @desc Builds DOM id attribute of the field,
 *       including layerId to namespace it
 *       and value in case if it's radio button,
 *       e.g. `signupForm___provider___facebook`
 *
 */
function buildFieldDomIdWithValue(layerId, userKeyPath, value) {
  if (!isString(value) && !isNumber(value)) {
    throw new Error(buildErrorMessage({
      layerId: layerId,
      message: ['Value for dom id must be a string or a number.', value ? 'Passed value is: ' + value : null, 'Field key path: ' + JSON.stringify(userKeyPath)]
    }));
  }

  var normalizedKeyPath = normalizeKeyPath(userKeyPath);

  return [layerId].concat(normalizedKeyPath).concat(value).join(Constant.FIELD_DOM_ID_DELIMITER);
}

/**
 * @desc Builds internal identificator of the slice of the state,
 *       that holds DOM props for the field,
 *       e.g. `fieldPropsState---user.email`
 *
 */
function buildFieldPropsStateId(fieldId) {
  var prefix = Constant.FIELD_PROPS_STATE_ID_PREFIX;
  var _ = Constant.FIELD_STATE_ID_DELIMITER;

  return '' + prefix + _ + fieldId;
}

/**
 * @desc Builds internal identificator of the slice of the state,
 *       that holds validation results for the field,
 *       e.g. `fieldValidationState---user.email`
 *
 */
function buildFieldValidationStateId(fieldId) {
  var prefix = Constant.FIELD_VALIDATION_STATE_ID_PREFIX;
  var _ = Constant.FIELD_STATE_ID_DELIMITER;

  return '' + prefix + _ + fieldId;
}

/**
 * @desc Parses internal identificator of the slice of the state,
 *       and returns type of data & fieldId,
 *       e.g. `fieldValidationState` & `user.email`
 *
 */
function parseFieldStateId(fieldStateId) {
  var _fieldStateId$split = fieldStateId.split(Constant.FIELD_STATE_ID_DELIMITER),
      dataType = _fieldStateId$split[0],
      fieldId = _fieldStateId$split[1];

  return { dataType: dataType, fieldId: fieldId };
}

/**
 * @desc On mount field validation.
 *       Results will be emitted on mount -> on every change.
 *
 */


var onMount = function onMount(field, value, data, event, stateContainer) {
  return buildCompleteSyncValidationResults(field, value, data, stateContainer.getPropsLevelStatuses(), stateContainer.getLayerId());
};

/**
 * @desc According to this strategy Layer won't emit any results
 *       until user types something in in a field.
 *       After first results are emitted-feedback is provided on every change.
 *
 */


var onFirstChange = function onFirstChange(field, value, data, event, stateContainer) {
  stateContainer.setEmittedField(field.id);
  return buildCompleteSyncValidationResults(field, value, data, stateContainer.getPropsLevelStatuses(), stateContainer.getLayerId());
};

/**
 * @desc According to this strategy Layer won't emit any results
 *       until user blures away from a field.
 *       After first results are emitted-feedback is provided on every change.
 *
 */
var onFirstBlur = function onFirstBlur(field, value, data, event, stateContainer) {
  var validationState = buildCompleteSyncValidationResults(field, value, data, stateContainer.getPropsLevelStatuses(), stateContainer.getLayerId());

  if (event && isBlurEvent(event)) {
    stateContainer.setBluredField(field.id);
    stateContainer.setEmittedField(field.id);
    return validationState;
  }

  if (stateContainer.getEmittedField(field.id)) {
    return validationState;
  }

  return buildEmptyValidationResults();
};

/**
 * @desc According to this strategy Layer won't emit any results
 *       until user types in valid value.
 *       After first results are emitted-feedback is provided on every change.
 *
 */


var onFirstSuccess = function onFirstSuccess(field, value, data, event, stateContainer) {
  var validationState = buildCompleteSyncValidationResults(field, value, data, stateContainer.getPropsLevelStatuses(), stateContainer.getLayerId());

  if (stateContainer.getEmittedField(field.id)) {
    return validationState;
  }

  if (validationState.valid) {
    stateContainer.setEmittedField(field.id);
    return validationState;
  }

  return buildEmptyValidationResults();
};

/**
 * @desc According to this strategy Layer won't emit any results
 *       until user types in valid value or blures away from a field.
 *       After first results are emitted-feedback is provided on every change.
 *
 */
var onFirstSuccessOrFirstBlur = function onFirstSuccessOrFirstBlur(field, value, data, event, stateContainer) {
  var validationState = buildCompleteSyncValidationResults(field, value, data, stateContainer.getPropsLevelStatuses(), stateContainer.getLayerId());

  if (event && isBlurEvent(event)) {
    stateContainer.setBluredField(field.id);
    stateContainer.setEmittedField(field.id);
    return validationState;
  }

  if (stateContainer.getEmittedField(field.id)) {
    return validationState;
  }

  if (validationState.valid) {
    stateContainer.setEmittedField(field.id);
    return validationState;
  }

  return buildEmptyValidationResults();
};

/**
 * @desc According to this strategy Layer won't emit any results
 *       until user submits the form.
 *       After first results are emitted-feedback is provided on every change.
 *
 */


var onFirstSubmit = function onFirstSubmit(field, value, data, event, stateContainer) {
  if (stateContainer.getFormWasSubmitted()) {
    stateContainer.setEmittedField(field.id);
    return buildCompleteSyncValidationResults(field, value, data, stateContainer.getPropsLevelStatuses(), stateContainer.getLayerId());
  }

  return buildEmptyValidationResults();
};

var _strategies;

/* eslint-disable key-spacing */

var strategies = (_strategies = {}, _strategies[Strategy.ON_MOUNT] = onMount, _strategies[Strategy.ON_FIRST_CHANGE] = onFirstChange, _strategies[Strategy.ON_FIRST_BLUR] = onFirstBlur, _strategies[Strategy.ON_FIRST_SUCCESS] = onFirstSuccess, _strategies[Strategy.ON_FIRST_SUCCESS_OR_FIRST_BLUR] = onFirstSuccessOrFirstBlur, _strategies[Strategy.ON_FIRST_SUBMIT] = onFirstSubmit, _strategies);

/**
 * @desc Async validation will be triggered on each blur event.
 *
 */


var onBlur = function onBlur(field, value, stateContainer) {
  stateContainer.setBluredField(field.id);
  stateContainer.setEmittedField(field.id);

  // [For ]: we're making sure it's fine in async validation method
  if (!field.validateAsync) throw new Error();

  // $FlowFixMe: No idea yet how to handle `async | debounced-async` case
  return field.validateAsync(value).then(function (results) {
    return {
      resolution: buildCompleteAsyncValidationResults(field.id, results, stateContainer.getPropsLevelStatuses(), stateContainer.getLayerId()),
      forValue: value
    };
  });
};

/**
 * @desc Field with this strategy has debounced async validators.
 *       When user starts typing, validator won't be triggered,
 *       but debouncer will report, that wait period has started
 *       or invokation still debounced. So this function returns
 *       only sync reports from debouncer, rather than Promise.
 *       When validator will be invoked from within debouncer,
 *       debouncer will notify state container directly.
 *
 */
var onChange = function onChange(field, value, stateContainer) {
  stateContainer.setEmittedField(field.id);

  // [For ]: we're making sure it's fine in async validation method
  if (!field.validateAsync) throw new Error();

  // $FlowFixMe: No idea yet how to handle `async | debounced-async` case
  return field.validateAsync(field.id, value, stateContainer);
};

var _asyncStrategies;

/* eslint-disable key-spacing */

var asyncStrategies = (_asyncStrategies = {}, _asyncStrategies[AsyncStrategy.ON_BLUR] = onBlur, _asyncStrategies[AsyncStrategy.ON_CHANGE] = onChange, _asyncStrategies);

/**
 * @desc Performs synchronous single field validation.
 *       It includes calculation of the sync strategy for the field.
 *
 */
function performSyncFieldValidation(field, value, data, event, stateContainer) {
  var ON_FIRST_BLUR = Strategy.ON_FIRST_BLUR,
      ON_FIRST_SUCCESS_OR_FIRST_BLUR = Strategy.ON_FIRST_SUCCESS_OR_FIRST_BLUR,
      DEFAULT = Strategy.DEFAULT;

  // Figuring out defined strategy

  var strategy = field.strategy || stateContainer.getPropsLevelStrategy() || DEFAULT;

  // Figuring out defined async strategy
  var asyncStrategy = field.asyncStrategy || stateContainer.getPropsLevelAsyncStrategy() || AsyncStrategy.DEFAULT;

  // Ignoring blur event, if strategies don't care about blur
  var blurConcernedStrategy = [ON_FIRST_BLUR, ON_FIRST_SUCCESS_OR_FIRST_BLUR].includes(strategy);
  var blurConcernedAsyncStrategy = !!field.validateAsync && asyncStrategy === AsyncStrategy.ON_BLUR;

  if (event && isBlurEvent(event) && !blurConcernedStrategy && !blurConcernedAsyncStrategy) {
    return null;
  }

  // Also ignoring blur event, if:
  //   - field doesn't have async validation w/ ON_BLUR strategy
  //   - and sync strategy is concerned about the first blur,
  //     but it's not the first blur event for this field
  if (event && isBlurEvent(event) && blurConcernedStrategy && !blurConcernedAsyncStrategy && (stateContainer.getBluredField(field.id) || stateContainer.getEmittedField(field.id))) {
    return null;
  }

  // If form was subitted, then switching to ON_FIRST_SUBMIT strategy
  // Otherwise using provided strategy
  var validateWithStrategy = stateContainer.getFormWasSubmitted() ? strategies.onFirstSubmit : strategies[strategy];

  if (!validateWithStrategy) {
    throw new Error(buildErrorMessage({
      layerId: stateContainer.getLayerId(),
      fieldId: field.id,
      message: 'Unknown strategy: ' + strategy + '. Most likely it\'s a spelling issue.'
    }));
  }

  // Validating with the calculated strategy
  var validatedWithStrategy = validateWithStrategy(field, value, data, event, stateContainer);

  // Before emit one more case must be handled:
  // if this is `change` event and validation was successful,
  // but field has also async validation w/ ON_BLUR stategy.
  // In this case not emitting anything until the blur.
  return blurConcernedAsyncStrategy && validatedWithStrategy.valid && event && isChangeEvent(event) ? buildEmptyValidationResults() : validatedWithStrategy;
}

/**
 * @desc Performs asynchronous single field validation.
 *       It includes calculation of the async strategy for the field.
 *
 */
function performAsyncFieldValidation(syncValidationResults, field, value, event, stateContainer) {
  // Don't perform async validation if:
  //   - there is no async validator
  //   - value is empty (no reason to validate empty value on remote)
  //   - value is invalid localy (no reason to validate invalid value on remote)
  //   - sync validation said `don't-provide-feedback-yet`
  var noAsyncValidator = !field.validateAsync;
  var noValue = !value && value !== 0;
  var isSyncValidationInvalid = syncValidationResults && !syncValidationResults.valid;
  var noEmittableSyncResults = syncValidationResults && syncValidationResults.valid === null && syncValidationResults.status === null && syncValidationResults.message === null;

  if (noAsyncValidator || noValue || isSyncValidationInvalid || noEmittableSyncResults) {
    return syncValidationResults;
  }

  // Figuring out async strategy
  var asyncStrategy = field.asyncStrategy || stateContainer.getPropsLevelAsyncStrategy() || AsyncStrategy.DEFAULT;

  // If it's ON_BLUR and event.type is blur -> validate
  if (event && isBlurEvent(event) && asyncStrategy === AsyncStrategy.ON_BLUR) {
    return asyncStrategies.onBlur(field, value, stateContainer);
  }

  // Don't perform async validation if:
  //   - validation is skipped by sync validator (null value returned)
  //   - it's a change event, but async strategy is ON_BLUR
  var isSyncValidationSkipped = syncValidationResults === null;
  var isOnChangeEventButOnBlurStrategy = event && isChangeEvent(event) && asyncStrategy === AsyncStrategy.ON_BLUR;

  if (isSyncValidationSkipped || isOnChangeEventButOnBlurStrategy) {
    return syncValidationResults;
  }

  // Otherwise -> debounced ON_CHANGE
  return asyncStrategies.onChange(field, value, stateContainer);
}

/**
 * @desc Controller of sync & async validations.
 *
 */
function performFieldValidation(parentField, parentValue, parentEvent, stateContainer) {
  if (parentField.linkedFields && !Array.isArray(parentField.linkedFields)) {
    throw new Error(buildErrorMessage({
      layerId: stateContainer.getLayerId(),
      fieldId: parentField.id,
      message: '`linkedFields` must be an Array.'
    }));
  }

  // Figuring out set of fields to validate:
  //   - parent field (which's changed triggered validation)
  //   - linked fields (if any)
  //
  // NOTE: Theoretically linked field might has its own linked fields.
  //       Something to keep in mind, but ignoring until first use case.
  var fields = parentField.linkedFields ? parentField.linkedFields.reduce(function (allFields, keyPath) {
    var fieldId = buildFieldIdFromUserKeyPath(keyPath);
    return allFields.concat(stateContainer.getNormalizedField(fieldId));
  }, [parentField]) : [parentField];

  // As we (might) have multiple fields to validate,
  // we don't want to trigger state update for each field separately,
  // but we want to batch updates and set next state within single transaction:
  // set sync valiudation results + start processing async ones.
  // Later on, when async results will be resolved ->
  // state will be updated again for each resolved field separetly.
  // Sync state will be packed under `nextSyncState` key,
  // Promises w/ ongoing async validations - under `ongoingAsyncValidations` key.
  var origin = { nextSyncState: {}, ongoingAsyncValidations: {} };

  return fields.reduce(function (results, field) {
    var _babelHelpers$extends4;

    var isParentField = field.id === parentField.id;

    var currentData = stateContainer.getData();

    // $FlowIgnoreMe: We're making sure that value at keyPath is not an object on normalization
    var value = isParentField ? parentValue : getProp(currentData, field.keyPath);
    var event = isParentField ? parentEvent : null;

    // If it's a linked field, data must be updated w/ the next value of parent field.
    // Not doing this for parent field as it affects perf.
    var data = isParentField ? currentData : setProp(currentData, parentField.keyPath, parentValue);

    // Performing sync -> async validations
    var syncValidationResults = performSyncFieldValidation(field, value, data, event, stateContainer);
    var finalValidationResults = performAsyncFieldValidation(syncValidationResults, field, value, event, stateContainer);

    // If no results returned or it's a debounced results -> ignoring them
    if (!finalValidationResults || finalValidationResults.status === DebounceStatus.DEBOUNCED) {
      return results;
    }

    var stateId = buildFieldValidationStateId(field.id);

    // If it's a Promise, it means that we should set state of this field to processing
    // and store Promise to update validation state later on (in the core component)
    if (isPromise(finalValidationResults)) {
      var _babelHelpers$extends, _babelHelpers$extends2;

      return {
        nextSyncState: _extends({}, results.nextSyncState, (_babelHelpers$extends = {}, _babelHelpers$extends[stateId] = buildIntermediateAsyncValidationResults(), _babelHelpers$extends)),
        ongoingAsyncValidations: _extends({}, results.ongoingAsyncValidations, (_babelHelpers$extends2 = {}, _babelHelpers$extends2[stateId] = finalValidationResults, _babelHelpers$extends2))
      };
    }

    // If it's a message from the debouncer w/ the status INITIALIZED ->
    // just starting processing the field.
    // Debouncer will notify state container about triggered request directly
    // and pass Promise within notification
    if (finalValidationResults.status === DebounceStatus.INITIALIZED) {
      var _babelHelpers$extends3;

      return {
        nextSyncState: _extends({}, results.nextSyncState, (_babelHelpers$extends3 = {}, _babelHelpers$extends3[stateId] = buildIntermediateAsyncValidationResults(), _babelHelpers$extends3)),
        ongoingAsyncValidations: results.ongoingAsyncValidations
      };
    }

    // Otherwise it's a results of sync validation. Just merging them.
    return {
      nextSyncState: _extends({}, results.nextSyncState, (_babelHelpers$extends4 = {}, _babelHelpers$extends4[stateId] = finalValidationResults, _babelHelpers$extends4)),
      ongoingAsyncValidations: results.ongoingAsyncValidations
    };
  }, origin);
}

/**
 * @desc Performs fields validation on mount.
 *
 */
function performOnMountValidation(fields, stateContainer) {
  var layerId = stateContainer.getLayerId();
  var data = stateContainer.getData();
  var normalizedFields = stateContainer.getNormalizedFields();
  var propsLevelStatuses = stateContainer.getPropsLevelStatuses();
  var propsLevelStrategy = stateContainer.getPropsLevelStrategy();
  var propsLevelAsyncStrategy = stateContainer.getPropsLevelAsyncStrategy();

  var origin = { nextSyncState: {}, ongoingAsyncValidations: {} };

  return normalizedFields.filter(function (field) {
    return (field.strategy || propsLevelStrategy) === Strategy.ON_MOUNT;
  }).reduce(function (_ref, field) {
    var _babelHelpers$extends6, _babelHelpers$extends7;

    var nextSyncState = _ref.nextSyncState,
        ongoingAsyncValidations = _ref.ongoingAsyncValidations;

    stateContainer.setEmittedField(field.id);

    // $FlowIgnoreMe: We're making sure that value at keyPath is not an object on normalization
    var value = getProp(data, field.keyPath);
    var stateId = buildFieldValidationStateId(field.id);

    var fieldNextSyncValidationState = buildCompleteSyncValidationResults(field, value, data, propsLevelStatuses, layerId);

    // Don't perform async validation if:
    //   - there is no async validator
    //   - value is empty (no reason to validate empty value on remote)
    //   - value is invalid localy (no reason to validate invalid value on remote)
    var isAsync = !!field.validateAsync && (!!value || value === 0) && fieldNextSyncValidationState.valid;

    if (!isAsync) {
      var _babelHelpers$extends5;

      return {
        nextSyncState: _extends({}, nextSyncState, (_babelHelpers$extends5 = {}, _babelHelpers$extends5[stateId] = fieldNextSyncValidationState, _babelHelpers$extends5)),
        ongoingAsyncValidations: ongoingAsyncValidations
      };
    }

    var asyncStrategy = field.asyncStrategy || propsLevelAsyncStrategy || AsyncStrategy.DEFAULT;

    // If asyncStrategy is ON_CHANGE,
    // then validator on normalized field is debounced.
    // We'll use original one.
    var asyncValidator = asyncStrategy === AsyncStrategy.ON_CHANGE ? getProp(fields, field.keyPath).validateAsync : field.validateAsync;

    return {
      nextSyncState: _extends({}, nextSyncState, (_babelHelpers$extends6 = {}, _babelHelpers$extends6[stateId] = buildIntermediateAsyncValidationResults(), _babelHelpers$extends6)),
      ongoingAsyncValidations: _extends({}, ongoingAsyncValidations, (_babelHelpers$extends7 = {}, _babelHelpers$extends7[stateId] = asyncValidator(value).then(function (results) {
        return {
          resolution: buildCompleteAsyncValidationResults(field.id, results, propsLevelStatuses, layerId),
          forValue: value
        };
      }), _babelHelpers$extends7))
    };
  }, origin);
}

/**
 * @desc Performs fields validation on form submission.
 *       We don't want to perform async validations here
 *       as those will be performed on the server anyway - all within single request.
 *       However we should respect the current state of async validations.
 *       That being said, algoritm looks like this:
 *         - perform batch synchronous validation of all fields
 *         - resolve merge conflicts w/ existing state using the following regulations:
 *           - no current state -> just merge-in next one
 *           - current failure & next failure -> take next one
 *           - current success & next failure -> take next one
 *           - current failure & next success -> current isAsync ? keep current : take next one
 *           - current success & next success -> current isAsync ? keep current : take next one
 *         - return next composite state + validity resolution
 *
 */
function performOnSubmitValidation(data, state, fields, statuses, layerId) {
  var origin = { validationState: {}, isValid: true };

  return fields.reduce(function (_ref2, field) {
    var _babelHelpers$extends8;

    var validationState = _ref2.validationState,
        isValid = _ref2.isValid;

    // $FlowIgnoreMe: We're making sure that value at keyPath is not an object on normalization
    var value = getProp(data, field.keyPath);
    var fieldValidationStateId = buildFieldValidationStateId(field.id);

    var fieldCurrentValidationState = state[fieldValidationStateId];
    var fieldNextSyncValidationState = buildCompleteSyncValidationResults(field, value, data, statuses, layerId);

    var fieldNextValidationState = fieldCurrentValidationState && fieldNextSyncValidationState.valid && fieldCurrentValidationState.isAsync ? fieldCurrentValidationState : fieldNextSyncValidationState;

    return {
      isValid: !isValid ? isValid : !!fieldNextValidationState.valid,
      validationState: _extends({}, validationState, (_babelHelpers$extends8 = {}, _babelHelpers$extends8[fieldValidationStateId] = fieldNextValidationState, _babelHelpers$extends8))
    };
  }, origin);
}

/**
 * @desc Validation Layer accepts fields as Object w/ props
 *       (struture of this Object replicates the structure of the data object).
 *       For internal usage this Object is normalized to flat Array of fields.
 *       Also it wraps async validators with ON_CHANGE strategy in debouncer.
 *
 */
function normalizeFieldsFromProps(propsAsyncStrategy, propsDebounceInterval, fields, data) {
  var normalizedFields = arguments.length > 4 && arguments[4] !== undefined ? arguments[4] : [];
  var parentKeyPath = arguments.length > 5 && arguments[5] !== undefined ? arguments[5] : [];

  var keys = Object.keys(fields);

  var index = -1;
  // eslint-disable-next-line no-plusplus
  while (++index < keys.length) {
    var key = keys[index];
    var fieldsBranch = fields[key];
    var dataBranch = getProp(data, key);
    var keyPath = parentKeyPath.concat(key);

    if (isPlainObject(dataBranch)) {
      normalizeFieldsFromProps(propsAsyncStrategy, propsDebounceInterval, fieldsBranch, dataBranch, normalizedFields, keyPath);
    } else {
      var id = buildFieldId(keyPath);

      if (isPlainObject(fieldsBranch)) {
        var fieldAsyncStrategy = propsAsyncStrategy || fieldsBranch.asyncStrategy || AsyncStrategy.DEFAULT;
        var debounceInterval = propsDebounceInterval || fieldsBranch.debounceInterval || Constant.DEFAULT_DEBOUNCE_INTERVAL;
        var isDebouncedAsyncValidation = fieldsBranch.validateAsync && fieldAsyncStrategy === AsyncStrategy.ON_CHANGE;

        var normalizedField = isDebouncedAsyncValidation ? _extends({}, fieldsBranch, {
          id: id,
          keyPath: keyPath,
          validateAsync: debounce(fieldsBranch.validateAsync, debounceInterval)
        }) : _extends({}, fieldsBranch, {
          id: id,
          keyPath: keyPath
        });
        normalizedFields.push(normalizedField);
      } else {
        normalizedFields.push({ id: id, keyPath: keyPath });
      }
    }
  }

  return normalizedFields;
}

/**
 * @desc Normalizes errors, received from `submitCallbacks.onFailure`.
 *       It happens when API rejected submission and responded w/ errors.
 *
 */
function normalizeExternalErrors(errors) {
  var normalizedErrors = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : [];
  var parentKeyPath = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : [];

  var keys = Object.keys(errors);

  var index = -1;
  // eslint-disable-next-line no-plusplus
  while (++index < keys.length) {
    var key = keys[index];
    var errorsBranch = errors[key];
    var keyPath = parentKeyPath.concat(key);

    if (isPlainObject(errorsBranch)) {
      normalizeExternalErrors(errorsBranch, normalizedErrors, keyPath);
    } else {
      normalizedErrors.push({
        fieldId: buildFieldId(keyPath),
        message: errorsBranch
      });
    }
  }

  return normalizedErrors;
}

/**
 * @desc It is possible to define DOM handler on the field level
 *       as well as on the props level for all the fields.
 *       This function figures out which one to use for specific field.
 *
 */
function getFieldDomHandler(layerId, fieldId, handlerKey, fieldLevelDomHandlers, propsLevelDomHandlers) {
  var fieldLevelHandler = fieldLevelDomHandlers && fieldLevelDomHandlers[handlerKey];
  var propsLevelHandler = propsLevelDomHandlers && propsLevelDomHandlers[handlerKey];

  var handler = fieldLevelHandler || propsLevelHandler;

  if (handlerKey === 'onChange' && !handler) {
    throw new Error(buildErrorMessage({
      layerId: layerId,
      fieldId: fieldId,
      message: 'Looks like you forgot to provide `' + handlerKey + '` handler.'
    }));
  }

  if (handler && !isFunction(handler)) {
    throw new Error(buildErrorMessage({
      layerId: layerId,
      fieldId: fieldLevelHandler ? fieldId : null,
      message: '`' + handlerKey + '` must be a function.'
    }));
  }

  return handler;
}

/**
 * @desc Same as function above, but for value handlers,
 *       i.e. `filter`, `transformBeforeStore` & `transformBeforeRender`.
 *
 * NOTE: Might worth to remove this and allow only `field`-level definitions.
 *       Not documenting it for now.
 *
 */
function getFieldValueHandler(layerId, fieldId, handlerKey, fieldLevelValueHandler, propsLevelValueHandler) {
  var handler = fieldLevelValueHandler || propsLevelValueHandler;

  if (handler && !isFunction(handler)) {
    throw new Error(buildErrorMessage({
      layerId: layerId,
      fieldId: fieldLevelValueHandler ? fieldId : null,
      message: '`' + handlerKey + '` must be a function.'
    }));
  }

  return handler;
}

/**
 * @desc On each update layer calculates DOM props for each field and stores it in state.
 *       This function calcultates next slice of the state with the fields DOM props.
 *
 */
function getFieldsPropsState(layerId, normalizedFields, data, propsLevelTransformBeforeRender, isSubmitting) {
  return normalizedFields.reduce(function (fields, field) {
    var _fieldPropsStateId, _babelHelpers$extends;

    var fieldDomId = buildFieldDomId(layerId, field.keyPath);
    var fieldPropsStateId = buildFieldPropsStateId(field.id);

    // $FlowIgnoreMe: We're making sure that value at keyPath is not an object on normalization
    var fieldValue = getProp(data, field.keyPath);
    var fieldNormalizedDomValue = normalizeValueForDom(fieldValue);

    var transformBeforeRender = getFieldValueHandler(layerId, field.id, 'transformBeforeRender', field.transformBeforeRender, propsLevelTransformBeforeRender);

    var fieldDomValue = transformBeforeRender ? transformBeforeRender(fieldNormalizedDomValue, data) : fieldNormalizedDomValue;

    var isDisabled = isSubmitting || field.disabled;

    return _extends({}, fields, (_babelHelpers$extends = {}, _babelHelpers$extends[fieldPropsStateId] = (_fieldPropsStateId = {}, _fieldPropsStateId[Constant.FIELD_ID_DOM_DATA_ATTRIBUTE] = field.id, _fieldPropsStateId.id = fieldDomId, _fieldPropsStateId.name = field.id, _fieldPropsStateId.value = fieldDomValue, _fieldPropsStateId.disabled = isDisabled, _fieldPropsStateId), _babelHelpers$extends));
  }, {});
}

/**
 * @desc When DOM handler is triggered from the view
 *       this function takes event (and some additional data)
 *       and normalizes it to pass to validators & user handlers.
 *
 */
function getNextDataFromDom(layerId, data, event) {
  var domEl = event.target;

  var value = domEl.value,
      checked = domEl.checked;


  var fieldId = domEl.getAttribute(Constant.FIELD_ID_DOM_DATA_ATTRIBUTE);

  if (!fieldId) {
    throw new Error(buildErrorMessage({
      layerId: layerId,
      message: "Can't find field id."
    }));
  }

  var _parseFieldId = parseFieldId(fieldId),
      attr = _parseFieldId.attr,
      keyPath = _parseFieldId.keyPath;

  return {
    fieldId: fieldId,
    attr: attr,
    keyPath: keyPath,
    value: value,
    checked: checked,
    data: data,
    event: event
  };
}

/**
 * @desc Same as function above except it normalizes the data
 *       received from custom handler.
 *
 */
function getNextDataFromCustomHandler(fieldId, value, data, event) {
  var _parseFieldId2 = parseFieldId(fieldId),
      attr = _parseFieldId2.attr,
      keyPath = _parseFieldId2.keyPath;

  return {
    fieldId: fieldId,
    attr: attr,
    keyPath: keyPath,
    value: value,
    data: data,
    event: event
  };
}

/**
 * @desc Interface to access validation layer data from the views.
 *
 */

var LayerInterface = function LayerInterface(stateContainer, handlers) {
  var _this = this;

  classCallCheck(this, LayerInterface);

  this.getPropsFor = function (keyPath) {
    var field = _this.__getField(keyPath, 'getPropsFor');

    // $FlowIssue: exact type + destructuring: https://github.com/facebook/flow/issues/2405
    return _extends({}, field.props, {
      onBlur: _this.__handleDomBlur,
      onChange: _this.__handleDomChange
    });
  };

  this.getCheckboxPropsFor = function (keyPath) {
    var field = _this.__getField(keyPath, 'getCheckboxPropsFor');

    // $FlowIssue: exact type + destructuring: https://github.com/facebook/flow/issues/2405
    return _extends({}, field.props, {
      checked: !!field.props.value,
      onBlur: _this.__handleDomBlur,
      onChange: _this.__handleDomChange
    });
  };

  this.getRadioButtonPropsFor = function (keyPath, value) {
    var layerId = _this.__layerId;
    var field = _this.__getField(keyPath, 'getRadioButtonPropsFor');

    // $FlowIssue: exact type + destructuring: https://github.com/facebook/flow/issues/2405
    return _extends({}, field.props, {
      value: value,
      id: buildFieldDomIdWithValue(layerId, keyPath, value),
      checked: field.props.value === value,
      onBlur: _this.__handleDomBlur,
      onChange: _this.__handleDomChange
    });
  };

  this.getCustomPropsFor = function (keyPath, _ref) {
    var value = _ref.value,
        disabled = _ref.disabled,
        getChecked = _ref.getChecked;

    var layerId = _this.__layerId;
    var field = _this.__getField(keyPath, 'getCustomPropsFor');

    var customProps = _extends({}, field.props, {
      onBlur: _this.__handleDomBlur,
      onChange: _this.__handleDomChange
    });

    if (value) {
      customProps.id = buildFieldDomIdWithValue(layerId, keyPath, value);
    }

    if (disabled) {
      customProps.disabled = disabled;
    }

    if (getChecked) {
      customProps.checked = getChecked(field.props.value);
    }

    // $FlowIssue: exact type + destructuring: https://github.com/facebook/flow/issues/2405
    return customProps;
  };

  this.getSubmitButtonProps = function () {
    return {
      type: 'submit',
      disabled: _this.__isSubmitting
    };
  };

  this.getValidityFor = function (keyPath) {
    var field = _this.__getField(keyPath, 'getValidityFor');
    return field.resolution && typeof field.resolution.valid !== 'undefined' ? field.resolution.valid : null;
  };

  this.getStatusFor = function (keyPath) {
    var field = _this.__getField(keyPath, 'getStatusFor');
    return field.resolution && field.resolution.status ? field.resolution.status : null;
  };

  this.getMessageFor = function (keyPath) {
    var field = _this.__getField(keyPath, 'getMessageFor');
    return field.resolution && field.resolution.message ? field.resolution.message : null;
  };

  this.getAsyncStatusFor = function (keyPath) {
    var field = _this.__getField(keyPath, 'getAsyncStatusFor');
    return field.resolution && typeof field.resolution.isProcessing !== 'undefined' ? field.resolution.isProcessing : false;
  };

  this.getSubmissionStatus = function () {
    return _this.__isSubmitting;
  };

  this.getDomIdFor = function (keyPath, value) {
    var layerId = _this.__layerId;
    var field = _this.__getField(keyPath, 'getStatusFor');

    return value ? buildFieldDomIdWithValue(layerId, keyPath, value) : field.props.id;
  };

  this.getFieldIdFor = function (keyPath) {
    var field = _this.__getField(keyPath, 'getFieldIdFor');
    return field.props[Constant.FIELD_ID_DOM_DATA_ATTRIBUTE];
  };

  var state = stateContainer.getState();

  this.__layerId = stateContainer.getLayerId();

  this.__fields = Object.keys(state).reduce(function (fieldsState, fieldStateId) {
    var _babelHelpers$extends;

    var _parseFieldStateId = parseFieldStateId(fieldStateId),
        dataType = _parseFieldStateId.dataType,
        fieldId = _parseFieldStateId.fieldId;

    var nextFieldState = dataType === Constant.FIELD_PROPS_STATE_ID_PREFIX ? { props: state[fieldStateId] } : { resolution: state[fieldStateId] };

    return _extends({}, fieldsState, (_babelHelpers$extends = {}, _babelHelpers$extends[fieldId] = _extends({}, fieldsState[fieldId], nextFieldState), _babelHelpers$extends));
  }, {});

  this.__getField = function (keyPath, methodName) {
    var layerId = _this.__layerId;
    var fieldId = buildFieldIdFromUserKeyPath(keyPath);
    var field = _this.__fields[fieldId];

    if (!field) {
      throw new Error(buildErrorMessage({
        layerId: layerId,
        fieldId: fieldId,
        message: ['Can\'t find data for field at provided keyPath: ' + JSON.stringify(keyPath), 'Use string to fetch data for non-nested attributes: e.g. `layer.' + methodName + '(\'email\')`', 'Use array to fetch data for nested attributes: e.g. `layer.' + methodName + '([\'user\', \'email\'])`']
      }));
    }

    return field;
  };

  this.__isSubmitting = stateContainer.getIsSubmitting();

  this.__handleDomBlur = handlers.handleDomBlur;
  this.__handleDomChange = handlers.handleDomChange;

  this.notifyOnBlur = handlers.handleCustomBlur;
  this.notifyOnChange = handlers.handleCustomChange;
  this.handleSubmit = handlers.handleSubmit;
};

/**
 * @desc Stateless container of DOM event handlers:
 *         - onChange
 *         - onBlur
 *         - onSubmit
 *
 */

var LayerDomHandlers = function (_React$Component) {
  inherits(LayerDomHandlers, _React$Component);

  function LayerDomHandlers() {
    var _temp, _this, _ret;

    classCallCheck(this, LayerDomHandlers);

    for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }

    return _ret = (_temp = (_this = possibleConstructorReturn(this, _React$Component.call.apply(_React$Component, [this].concat(args))), _this), _this.handleDomChange = function (event) {
      var _this2 = _this,
          props = _this2.props;

      var domData = getNextDataFromDom(props.layerId, props.data, event);

      _this.handleChange(domData);
    }, _this.handleCustomChange = function (fieldId, value) {
      var _this3 = _this,
          props = _this3.props;

      var domData = getNextDataFromCustomHandler(fieldId, value, props.data);

      _this.handleChange(domData);
    }, _this.handleChange = function (domData) {
      var _this4 = _this,
          props = _this4.props;
      var stateContainer = props.stateContainer;


      var field = stateContainer.getNormalizedField(domData.fieldId);
      var filter = getFieldValueHandler(props.layerId, field.id, 'filter', field.filter, props.filter);

      if (filter // if filter is provided
      && domData.value // & there's a value is not an empty string
      && !filter(domData.value, props.data) // & filter function returned false
      ) {
          return; // then ignoring this update
        }

      var transformBeforeStore = getFieldValueHandler(props.layerId, field.id, 'transformBeforeStore', field.transformBeforeStore, props.transformBeforeStore);

      var processedValue = transformBeforeStore ? transformBeforeStore(domData.value, props.data) : domData.value;

      var processedDomData = _extends({}, domData, { value: processedValue });

      if (domData.event) {
        var handleChange = getFieldDomHandler(props.layerId, field.id, 'onChange', field.handlers, props.handlers);

        // $FlowIgnoreMe: Even getFieldDomHandler throws if onChange is missing, flow can't infer it
        handleChange(processedDomData);
      }

      stateContainer.validateField(field, processedDomData);
    }, _this.handleDomBlur = function (event) {
      var _this5 = _this,
          props = _this5.props;

      var domData = getNextDataFromDom(props.layerId, props.data, event);

      _this.handleBlur(domData);
    }, _this.handleCustomBlur = function (fieldId, value, event) {
      var _this6 = _this,
          props = _this6.props;

      var domData = getNextDataFromCustomHandler(fieldId, value, props.data, event);

      _this.handleBlur(domData, true);
    }, _this.handleBlur = function (domData) {
      var isCustom = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : false;
      var _this7 = _this,
          props = _this7.props;
      var stateContainer = props.stateContainer;


      var field = stateContainer.getNormalizedField(domData.fieldId);
      var handleBlur = getFieldDomHandler(props.layerId, field.id, 'onBlur', field.handlers, props.handlers);

      var transformBeforeStore = getFieldValueHandler(props.layerId, field.id, 'transformBeforeStore', field.transformBeforeStore, props.transformBeforeStore);

      var processedValue = transformBeforeStore ? transformBeforeStore(domData.value, props.data) : domData.value;

      var processedDomData = _extends({}, domData, { value: processedValue });

      if (!isCustom && handleBlur) handleBlur(processedDomData);

      stateContainer.validateField(field, processedDomData);
    }, _this.handleSubmit = function (event) {
      if (event && event.preventDefault) {
        event.preventDefault();
      }

      _this.props.stateContainer.triggerSubmission();
    }, _temp), possibleConstructorReturn(_this, _ret);
  }

  /**
   * @desc Change handlers.
   *
   */


  /**
   * @desc Blur handlers.
   *
   */


  /**
   * @desc Submit handler.
   *
   */


  LayerDomHandlers.prototype.render = function render() {
    var props = this.props;


    var layer = new LayerInterface(props.stateContainer, {
      handleDomBlur: this.handleDomBlur,
      handleDomChange: this.handleDomChange,
      handleCustomBlur: this.handleCustomBlur,
      handleCustomChange: this.handleCustomChange,
      handleSubmit: this.handleSubmit
    });

    if (typeof props.children !== 'function') {
      throw new Error(buildErrorMessage({
        layerId: props.layerId,
        message: '`children` must be a function'
      }));
    }

    return props.children(layer);
  };

  return LayerDomHandlers;
}(React__default.Component);

/**
 * @desc Core component. Container of the validation layer state.
 *
 */

var ValidationLayer = function (_Component) {
  inherits(ValidationLayer, _Component);

  function ValidationLayer(props) {
    classCallCheck(this, ValidationLayer);

    for (var _len = arguments.length, rest = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
      rest[_key - 1] = arguments[_key];
    }

    var _this = possibleConstructorReturn(this, _Component.call.apply(_Component, [this, props].concat(rest)));

    _initialiseProps.call(_this);

    _this.__layerId = props.id || Constant.DEFAULT_LAYER_ID;

    _this.__normalizedFields = normalizeFieldsFromProps(props.asyncStrategy, props.debounceInterval, props.fields, props.data);
    return _this;
  }

  /**
   * @desc Lifecycle hooks.
   *
   */


  ValidationLayer.prototype.componentWillMount = function componentWillMount() {
    // Set fields props
    this.setNextFieldsPropsState();
  };

  ValidationLayer.prototype.componentDidMount = function componentDidMount() {
    // Performs validation of fields w/ `onMount` strategy
    this.validateOnMount();
  };

  ValidationLayer.prototype.componentWillReceiveProps = function componentWillReceiveProps(nextProps) {
    // Update normalized fields
    this.setNextNormalizedFields(nextProps);

    // Update fields props
    this.setNextFieldsPropsState(nextProps);
  };

  /**
   * @desc Internal getters.
   *
   */


  /**
   * @desc Fields DOM props management.
   *
   */


  /**
   * @desc Normalized fields management.
   *
   */


  /**
   * @desc Blured fields management.
   *
   */


  /**
   * @desc Emitted fields management.
   *
   */


  /**
   * @desc Returns `true` if any ongoing async validation is happening.
   *
   */


  /**
   * @desc Get and set flag if form is submitting.
   *
   */


  /**
   * @desc Get and set flag if form was submitted at least once.
   *
   */


  /**
   * @desc On mount fields validation.
   *
   */


  /**
   * @desc Field validation on `change` & `blur` events.
   *       This method is triggered from <LayerDomHandlers />.
   *
   */


  /**
   * @desc Async validation notifier.
   *       This method is triggered from debouncer,
   *       when async validator is ready to run.
   *
   */


  /**
   * @desc Sets next validation state from various validators.
   *
   */


  /**
   * @desc Sets next deferred validation state from async validator
   *       (if it's not obsolete).
   *
   */


  /**
   * @desc Submission handler.
   *
   */


  /**
   * @desc Form validation on submission.
   *
   */


  /**
   * @desc Handler of form submission failure.
   *       This method is triggered from userland, when form submission failed.
   *       It accepts single argument: errors, reported from API (or anywhere else).
   *
   */


  /**
   * @desc State reset. Triggered from userland.
   *       Back to origins on successful form submission.
   *
   */


  ValidationLayer.prototype.render = function render() {
    var props = this.props;

    var layerId = this.getLayerId();

    if (typeof props.children !== 'function') {
      throw new Error(buildErrorMessage({
        layerId: layerId,
        message: '`children` must be a function'
      }));
    }

    return React__default.createElement(
      LayerDomHandlers,
      {
        layerId: layerId,
        stateContainer: this,
        data: props.data,
        filter: props.filter,
        handlers: props.handlers,
        transformBeforeStore: props.transformBeforeStore
      },
      function (layer) {
        return props.children && props.children(layer);
      }
    );
  };

  return ValidationLayer;
}(React.Component);

var _initialiseProps = function _initialiseProps() {
  var _this2 = this;

  this.state = {};
  this.__normalizedFields = [];
  this.__bluredFields = {};
  this.__emittedFields = {};
  this.__isSubmitting = false;
  this.__formWasSubmitted = false;

  this.getLayerId = function () {
    return _this2.__layerId;
  };

  this.getState = function () {
    return _this2.state;
  };

  this.getData = function () {
    return _this2.props.data;
  };

  this.getPropsLevelStrategy = function () {
    return _this2.props.strategy;
  };

  this.getPropsLevelAsyncStrategy = function () {
    return _this2.props.asyncStrategy;
  };

  this.getPropsLevelStatuses = function () {
    return _this2.props.statuses;
  };

  this.getNextFieldsPropsState = function (nextProps) {
    var props = nextProps || _this2.props;

    return getFieldsPropsState(_this2.getLayerId(), _this2.getNormalizedFields(), props.data, props.transformBeforeRender, _this2.getIsSubmitting());
  };

  this.setNextFieldsPropsState = function (nextProps, callback) {
    var nextState = _this2.getNextFieldsPropsState(nextProps);
    _this2.setState(nextState, callback);
  };

  this.getNormalizedFields = function () {
    return _this2.__normalizedFields;
  };

  this.getNormalizedField = function (fieldId) {
    var normalizedField = _this2.__normalizedFields.find(function (field) {
      return field.id === fieldId;
    });

    if (!normalizedField) {
      var layerId = _this2.getLayerId();

      throw new Error(buildErrorMessage({
        layerId: layerId,
        fieldId: fieldId,
        message: ['Can\'t find `field` at key path: ' + fieldId, 'Make sure it exists and has truthy value (e.g. object with options or just `true`).']
      }));
    }

    return normalizedField;
  };

  this.setNextNormalizedFields = function (nextProps) {
    var props = nextProps || _this2.props;

    var nextNormalizedFields = normalizeFieldsFromProps(props.asyncStrategy, props.debounceInterval, props.fields, props.data);

    // Adding only new fileds to prevent replacement
    // of debounced async validators
    var normalizedFields = nextNormalizedFields.map(function (nextNormalizedField) {
      var normalizedField = _this2.__normalizedFields.find(function (field) {
        return field.id === nextNormalizedField.id;
      });
      return normalizedField || nextNormalizedField;
    });

    _this2.__normalizedFields = normalizedFields;
  };

  this.getBluredField = function (fieldId) {
    return _this2.__bluredFields[fieldId];
  };

  this.setBluredField = function (fieldId) {
    if (!_this2.__bluredFields[fieldId]) {
      _this2.__bluredFields[fieldId] = true;
    }
  };

  this.getEmittedField = function (fieldId) {
    return _this2.__emittedFields[fieldId];
  };

  this.setEmittedField = function (fieldId) {
    if (!_this2.__emittedFields[fieldId]) {
      _this2.__emittedFields[fieldId] = true;
    }
  };

  this.setAllFieldsEmitted = function () {
    _this2.__emittedFields = _this2.__normalizedFields.map(function (field) {
      return field.id;
    }).reduce(function (fields, fieldId) {
      var _babelHelpers$extends;

      return _extends({}, fields, (_babelHelpers$extends = {}, _babelHelpers$extends[fieldId] = true, _babelHelpers$extends));
    }, {});
  };

  this.isAnyOngoingAsyncActivity = function () {
    var state = _this2.state;
    // $FlowIgnoreMe: Flow, don't be such an ass

    return Object.keys(state).some(function (stateId) {
      return state[stateId].isProcessing;
    });
  };

  this.getIsSubmitting = function () {
    return _this2.__isSubmitting;
  };

  this.setIsSubmitting = function (isSubmitting) {
    _this2.__isSubmitting = isSubmitting;
  };

  this.getFormWasSubmitted = function () {
    return _this2.__formWasSubmitted;
  };

  this.setFormWasSubmitted = function () {
    if (!_this2.__formWasSubmitted) {
      _this2.__formWasSubmitted = true;
    }
  };

  this.validateOnMount = function () {
    var fields = _this2.props.fields;

    var _performOnMountValida = performOnMountValidation(fields, _this2),
        nextSyncState = _performOnMountValida.nextSyncState,
        ongoingAsyncValidations = _performOnMountValida.ongoingAsyncValidations;

    _this2.setNextValidationState(nextSyncState, ongoingAsyncValidations);
  };

  this.validateField = function (field, domData) {
    var _performFieldValidati = performFieldValidation(field, domData.value, domData.event, _this2),
        nextSyncState = _performFieldValidati.nextSyncState,
        ongoingAsyncValidations = _performFieldValidati.ongoingAsyncValidations;

    _this2.setNextValidationState(nextSyncState, ongoingAsyncValidations);
  };

  this.notifyAsync = function (fieldId, value, asyncValidation) {
    var stateId = buildFieldValidationStateId(fieldId);

    asyncValidation.then(function (results) {
      var deferredNextState = {
        resolution: buildCompleteAsyncValidationResults(fieldId, results, _this2.getPropsLevelStatuses(), _this2.getLayerId()),
        forValue: value
      };

      _this2.setNextDeferredValidationState(stateId, deferredNextState);
    });
  };

  this.setNextValidationState = function (nextSyncState, ongoingAsyncValidations) {
    if (Object.keys(nextSyncState).length === 0) return;

    _this2.setState(nextSyncState, function () {
      Object.keys(ongoingAsyncValidations).forEach(function (stateId) {
        ongoingAsyncValidations[stateId].then(function (deferredNextState) {
          return _this2.setNextDeferredValidationState(stateId, deferredNextState);
        });
      });
    });
  };

  this.setNextDeferredValidationState = function (stateId, nextState) {
    var _this2$setState;

    if (!nextState || !nextState.resolution) return;

    var props = _this2.props;

    var _parseFieldStateId = parseFieldStateId(stateId),
        fieldId = _parseFieldStateId.fieldId;

    var _parseFieldId = parseFieldId(fieldId),
        keyPath = _parseFieldId.keyPath;

    var fieldValue = getProp(props.data, keyPath);

    // While async call is being processed,
    // deferred validation state might become obsolete:
    // i.e. value was changed & state was replaced
    // w/ next validation results for the next value
    if (nextState.forValue !== fieldValue) return;

    _this2.setState((_this2$setState = {}, _this2$setState[stateId] = nextState.resolution, _this2$setState));
  };

  this.triggerSubmission = function () {
    // Not submitting anything if there's any ongoing async activity
    // or form is submitting at the moment
    if (_this2.isAnyOngoingAsyncActivity() || _this2.getIsSubmitting()) {
      return;
    }

    _this2.setIsSubmitting(true);
    _this2.setFormWasSubmitted();
    _this2.setAllFieldsEmitted();
    _this2.setNextFieldsPropsState(null, _this2.validateForm);
  };

  this.validateForm = function () {
    var _performOnSubmitValid = performOnSubmitValidation(_this2.getData(), _this2.getState(), _this2.getNormalizedFields(), _this2.getPropsLevelStatuses(), _this2.getLayerId()),
        validationState = _performOnSubmitValid.validationState,
        isValid = _performOnSubmitValid.isValid;

    if (isValid) {
      var handlers = _this2.props.handlers;


      if (!handlers || !handlers.onSubmit || !isFunction(handlers.onSubmit)) {
        throw new Error(buildErrorMessage({
          layerId: _this2.getLayerId(),
          message: "Make sure `handlers.onSubmit` is exists and it's a function"
        }));
      }

      handlers.onSubmit({
        onSuccess: _this2.resetState,
        onFailure: _this2.handleFailurePostSubmitAction
      });
    } else {
      _this2.setState(validationState, _this2.handleFailurePostSubmitAction);
    }
  };

  this.handleFailurePostSubmitAction = function (errors) {
    _this2.setIsSubmitting(false);

    if (!errors) return _this2.setNextFieldsPropsState();

    var nextFieldsPropsState = _this2.getNextFieldsPropsState();

    var statuses = _this2.getPropsLevelStatuses();
    var normalizedErrors = normalizeExternalErrors(errors);
    var nextFieldsValidationState = normalizedErrors.reduce(function (nextValidationState, error) {
      var _babelHelpers$extends2;

      var stateId = buildFieldValidationStateId(error.fieldId);

      return _extends({}, nextValidationState, (_babelHelpers$extends2 = {}, _babelHelpers$extends2[stateId] = {
        valid: false,
        status: statuses && statuses.failure ? statuses.failure : DefaultStatus.FAILURE,
        message: error.message,
        isAsync: true
      }, _babelHelpers$extends2));
    }, {});

    var nextState = _extends({}, nextFieldsPropsState, nextFieldsValidationState);

    return _this2.setState(nextState);
  };

  this.resetState = function () {
    _this2.setNextNormalizedFields();

    _this2.__bluredFields = {};
    _this2.__emittedFields = {};

    _this2.__isSubmitting = false;
    _this2.__formWasSubmitted = false;

    // All values of the state keys must be set to an empty objects
    // to replace stale data, b/c setState() merges updates w/ current state.
    var resetedState = Object.keys(_this2.state).reduce(function (nextState, key) {
      var _babelHelpers$extends3;

      return _extends({}, nextState, (_babelHelpers$extends3 = {}, _babelHelpers$extends3[key] = {}, _babelHelpers$extends3));
    }, {});
    var nextFieldsPropsState = _this2.getNextFieldsPropsState();

    _this2.setState(_extends({}, resetedState, nextFieldsPropsState));
  };
};

exports['default'] = ValidationLayer;
exports.normalizeValidationResults = normalizeValidationResults;

Object.defineProperty(exports, '__esModule', { value: true });

})));
